"""App definition for reviewboard.accounts.

Version Added:
    5.0
"""

from django.apps import AppConfig


class AccountsAppConfig(AppConfig):
    """App configuration for reviewboard.accounts.

    Version Added
        5.0
    """

    name = 'reviewboard.accounts'

    def ready(self):
        """Configure the app once it's ready.

        This will populate the SSO backends registry.
        """
        from reviewboard.accounts.sso.backends import sso_backends
        sso_backends.populate()



"""Unit tests for reviewboard.datagrids.columns.MyCommentsColumn.

Version Added:
    8.0
"""

from __future__ import annotations

from django.contrib.auth.models import AnonymousUser
from django.utils.safestring import SafeString

from reviewboard.datagrids.columns import MyCommentsColumn
from reviewboard.datagrids.tests.base import BaseColumnTestCase


class MyCommentsColumnTests(BaseColumnTestCase):
    """Unit tests for reviewboard.datagrids.columns.MyCommentsColumn.

    Version Added:
        8.0
    """

    column = MyCommentsColumn()

    def test_render_data_as_anonymous(self) -> None:
        """Testing MyCommentsColumn.render_data as anonymous user"""
        review_request = self.create_review_request(publish=True)
        self.request.user = AnonymousUser()

        value = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(value, SafeString)
        self.assertEqual(value, '')

    def test_render_data_with_no_reviews(self) -> None:
        """Testing MyCommentsColumn.render_data with no reviews"""
        review_request = self.create_review_request(publish=True)

        # These are normally set by the column's augment_queryset().
        review_request.mycomments_my_reviews = 0
        review_request.mycomments_private_reviews = 0
        review_request.mycomments_shipit_reviews = 0

        value = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(value, SafeString)
        self.assertEqual(value, '')

    def test_render_data_with_published_review(self) -> None:
        """Testing MyCommentsColumn.render_data with published review"""
        review_request = self.create_review_request(publish=True)

        # These are normally set by the column's augment_queryset().
        review_request.mycomments_my_reviews = 1
        review_request.mycomments_private_reviews = 0
        review_request.mycomments_shipit_reviews = 0

        value = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(value, SafeString)
        self.assertHTMLEqual(
            value,
            '<div class="rb-icon rb-icon-datagrid-comment"'
            ' title="Comments published"></div>')

    def test_render_data_with_draft_review(self) -> None:
        """Testing MyCommentsColumn.render_data with draft review"""
        review_request = self.create_review_request(publish=True)

        # These are normally set by the column's augment_queryset().
        review_request.mycomments_my_reviews = 1
        review_request.mycomments_private_reviews = 1
        review_request.mycomments_shipit_reviews = 0

        value = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(value, SafeString)
        self.assertHTMLEqual(
            value,
            '<div class="rb-icon rb-icon-datagrid-comment-draft"'
            ' title="Comments drafted"></div>')

    def test_render_data_with_ship_it(self) -> None:
        """Testing MyCommentsColumn.render_data with Ship It"""
        review_request = self.create_review_request(publish=True)

        # These are normally set by the column's augment_queryset().
        review_request.mycomments_my_reviews = 1
        review_request.mycomments_private_reviews = 0
        review_request.mycomments_shipit_reviews = 1

        value = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(value, SafeString)
        self.assertHTMLEqual(
            value,
            '<div class="rb-icon rb-icon-datagrid-comment-shipit"'
            ' title="Comments published. Ship it!"></div>')

    def test_render_data_draft_prioritized_over_ship_it(self) -> None:
        """Testing MyCommentsColumn.render_data with draft prioritized over
        Ship It
        """
        review_request = self.create_review_request(publish=True)

        # These are normally set by the column's augment_queryset().
        review_request.mycomments_my_reviews = 2
        review_request.mycomments_private_reviews = 1
        review_request.mycomments_shipit_reviews = 1

        value = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(value, SafeString)
        self.assertHTMLEqual(
            value,
            '<div class="rb-icon rb-icon-datagrid-comment-draft"'
            ' title="Comments drafted"></div>')

    def test_to_json_as_anonymous(self) -> None:
        """Testing MyCommentsColumn.to_json as anonymous user"""
        review_request = self.create_review_request(publish=True)
        self.request.user = AnonymousUser()

        self.assertEqual(
            self.column.to_json(self.stateful_column, review_request),
            {})

    def test_to_json_with_no_reviews(self) -> None:
        """Testing MyCommentsColumn.to_json with no reviews"""
        review_request = self.create_review_request(publish=True)

        # These are normally set by the column's augment_queryset().
        review_request.mycomments_my_reviews = 0
        review_request.mycomments_private_reviews = 0
        review_request.mycomments_shipit_reviews = 0

        self.assertEqual(
            self.column.to_json(self.stateful_column, review_request),
            {})

    def test_to_json_with_published_review(self) -> None:
        """Testing MyCommentsColumn.to_json with published review"""
        review_request = self.create_review_request(publish=True)

        # These are normally set by the column's augment_queryset().
        review_request.mycomments_my_reviews = 1
        review_request.mycomments_private_reviews = 0
        review_request.mycomments_shipit_reviews = 0

        self.assertEqual(
            self.column.to_json(self.stateful_column, review_request),
            {
                'has_draft_reviews': False,
                'has_ship_its': False,
                'status': 'has-review',
            })

    def test_to_json_with_draft_review(self) -> None:
        """Testing MyCommentsColumn.to_json with draft review"""
        review_request = self.create_review_request(publish=True)

        # These are normally set by the column's augment_queryset().
        review_request.mycomments_my_reviews = 1
        review_request.mycomments_private_reviews = 1
        review_request.mycomments_shipit_reviews = 0

        self.assertEqual(
            self.column.to_json(self.stateful_column, review_request),
            {
                'has_draft_reviews': True,
                'has_ship_its': False,
                'status': 'has-draft-review',
            })

    def test_to_json_with_ship_it(self) -> None:
        """Testing MyCommentsColumn.to_json with Ship It"""
        review_request = self.create_review_request(publish=True)

        # These are normally set by the column's augment_queryset().
        review_request.mycomments_my_reviews = 1
        review_request.mycomments_private_reviews = 0
        review_request.mycomments_shipit_reviews = 1

        self.assertEqual(
            self.column.to_json(self.stateful_column, review_request),
            {
                'has_draft_reviews': False,
                'has_ship_its': True,
                'status': 'has-ship-it',
            })



"""Registry for managing available hosting services.

Version Added:
    6.0:
    This replaces the registry code in the old
    :py:mod:`reviewboard.hostingsvcs.service` module.
"""

from __future__ import annotations

import inspect
import logging
import re
from importlib import import_module
from typing import Type, TYPE_CHECKING

from django.urls import include, re_path
from django.utils.translation import gettext_lazy as _
from djblets.registries.errors import ItemLookupError
from djblets.registries.registry import (ALREADY_REGISTERED, LOAD_ENTRY_POINT,
                                         NOT_REGISTERED)

from reviewboard.hostingsvcs.urls import dynamic_urls as hostingsvcs_urls
from reviewboard.hostingsvcs.base.hosting_service import BaseHostingService
from reviewboard.registries.registry import EntryPointRegistry

if TYPE_CHECKING:
    from collections.abc import Iterator, Sequence
    from importlib.metadata import EntryPoint

    from django.urls import _AnyURL


logger = logging.getLogger(__name__)


class HostingServiceRegistry(EntryPointRegistry[Type[BaseHostingService]]):
    """A registry for managing hosting services.

    Version Changed:
        6.0:
        * Moved from :py:mod:`reviewboard.hostingsvcs.service` to
          :py:mod:`reviewboard.hostingsvcs.base.registry`.
    """

    entry_point = 'reviewboard.hosting_services'
    lookup_attrs = ['hosting_service_id']

    errors = {
        ALREADY_REGISTERED: _(
            '"%(item)s" is already a registered hosting service.'
        ),
        LOAD_ENTRY_POINT: _(
            'Unable to load repository hosting service %(entry_point)s: '
            '%(error)s.'
        ),
        NOT_REGISTERED: _(
            '"%(attr_value)s" is not a registered hosting service.'
        ),
    }

    ######################
    # Instance variables #
    ######################

    #: A mapping of hosting service IDs to URL patterns/resolvers.
    #:
    #: Type:
    #:     dict
    _url_patterns: dict[str, Sequence[_AnyURL]]

    def __init__(self) -> None:
        """Initialize the registry."""
        super().__init__()

        self._url_patterns = {}

    def get_defaults(self) -> Iterator[type[BaseHostingService]]:
        """Yield the built-in hosting services.

        This will make sure the standard hosting services are always present in
        the registry.

        Yields:
            type:
            The :py:class:`~reviewboard.hostingsvcs.base.hosting_service.
            BaseHostingService` subclasses.
        """
        for _module, _service_cls_name in (
                ('assembla', 'Assembla'),
                ('beanstalk', 'Beanstalk'),
                ('bitbucket', 'Bitbucket'),
                ('bugzilla', 'Bugzilla'),
                ('codebasehq', 'CodebaseHQ'),
                ('fedorahosted', 'FedoraHosted'),
                ('fogbugz', 'FogBugz'),
                ('forgejo', 'Forgejo'),
                ('gerrit', 'Gerrit'),
                ('github', 'GitHub'),
                ('gitlab', 'GitLab'),
                ('gitorious', 'Gitorious'),
                ('googlecode', 'GoogleCode'),
                ('jira', 'JIRA'),
                ('kiln', 'Kiln'),
                ('rbgateway', 'ReviewBoardGateway'),
                ('redmine', 'Redmine'),
                ('sourceforge', 'SourceForge'),
                ('splat', 'Splat'),
                ('trac', 'Trac'),
                ('unfuddle', 'Unfuddle'),
                ('versionone', 'VersionOne'),
            ):
            mod = import_module(f'reviewboard.hostingsvcs.{_module}')

            yield getattr(mod, _service_cls_name)

        yield from super().get_defaults()

    def get_hosting_service(
        self,
        hosting_service_id: str,
    ) -> type[BaseHostingService] | None:
        """Return a hosting service with the given ID.

        Args:
            hosting_service_id (str):
                The hosting service ID to return.

        Returns:
            type:
            The hosting service class, or ``None`` if not found.
        """
        try:
            return self.get('hosting_service_id', hosting_service_id)
        except ItemLookupError:
            return None

    def unregister(
        self,
        service: type[BaseHostingService],
    ) -> None:
        """Unregister a hosting service.

        This will also remove all registered URLs that the hosting service has
        defined.

        Args:
            service (type):
                The
                :py:class:`~reviewboard.hostingsvcs.base.hosting_service.
                BaseHostingService` subclass.
        """
        hosting_service_id = service.hosting_service_id

        super().unregister(service)

        if hosting_service_id and hosting_service_id in self._url_patterns:
            cls_urlpatterns = self._url_patterns[hosting_service_id]
            hostingsvcs_urls.remove_patterns(cls_urlpatterns)
            del self._url_patterns[hosting_service_id]

    def unregister_by_id(
        self,
        hosting_service_id: str,
    ) -> None:
        """Unregister a hosting service by ID.

        This will also remove all registered URLs that the hosting service has
        defined.

        Args:
            hosting_service_id (str):
                The ID of the hosting service to unregister.
        """
        try:
            self.unregister_by_attr('hosting_service_id', hosting_service_id)
        except ItemLookupError:
            logger.error('Failed to unregister unknown hosting service "%s"',
                         hosting_service_id)

            raise

    def process_value_from_entry_point(
        self,
        entry_point: EntryPoint,
    ) -> type[BaseHostingService]:
        """Load the class from the entry point.

        The ``hosting_service_id`` attribute will be set on the class from the
        entry point's name.

        Args:
            entry_point (importlib.metadata.EntryPoint):
                The entry point.

        Returns:
            type:
            The :py:class:`HostingService` subclass.
        """
        cls = entry_point.load()
        assert inspect.isclass(cls)
        assert issubclass(cls, BaseHostingService)

        cls.hosting_service_id = entry_point.name
        return cls

    def register(
        self,
        service: type[BaseHostingService],
    ) -> None:
        """Register a hosting service.

        This also adds the URL patterns defined by the hosting service. If the
        hosting service has a :py:attr:`HostingService.repository_url_patterns`
        attribute that is non-``None``, they will be automatically added.

        Args:
            service (type):
                The :py:class:`HostingService` subclass.
        """
        super().register(service)

        if service.repository_url_patterns:
            assert service.hosting_service_id

            escaped_id = re.escape(service.hosting_service_id)

            cls_urlpatterns = [
                re_path(rf'^(?P<hosting_service_id>{escaped_id})/',
                        include(service.repository_url_patterns)),
            ]

            self._url_patterns[service.hosting_service_id] = cls_urlpatterns
            hostingsvcs_urls.add_patterns(cls_urlpatterns)


#: The main registry of hosting services.
#:
#: Version Added:
#:     6.0
#:
#: Type:
#:     HostingServiceRegistry
hosting_service_registry = HostingServiceRegistry()



"""SCMTool for Mercurial."""

from __future__ import annotations

import json
import logging
from datetime import datetime
from typing import Any, Optional, TYPE_CHECKING, Union
from urllib.parse import quote as urllib_quote, urlparse

from django.utils.encoding import force_str
from djblets.util.filesystem import is_exe_in_path

from reviewboard.diffviewer.parser import DiffParser, DiffParserError
from reviewboard.scmtools.core import (
    Branch,
    Commit,
    HEAD,
    PRE_CREATION,
    SCMClient,
    SCMTool,
    UNKNOWN,
)
from reviewboard.scmtools.errors import FileNotFoundError, SCMError
from reviewboard.scmtools.git import GitDiffParser, strip_git_symlink_mode

if TYPE_CHECKING:
    import subprocess

    from reviewboard.diffviewer.parser import ParsedDiffFile
    from reviewboard.scmtools.core import (
        FileLookupContext,
        Revision,
        RevisionID,
    )
    from reviewboard.scmtools.models import Repository
    from reviewboard.site.models import LocalSite


logger = logging.getLogger(__name__)


INITIAL_COMMIT_ID = '0' * 40


class HgTool(SCMTool):
    """SCMTool for Mercurial."""

    scmtool_id = 'mercurial'
    name = "Mercurial"
    diffs_use_absolute_paths = True
    supports_history = True
    supports_post_commit = True
    dependencies = {
        'executables': ['hg'],
    }

    def __init__(
        self,
        repository: Repository,
    ) -> None:
        """Initialize the SCMTool.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository owning the instance of this tool.
        """
        super().__init__(repository)

        if repository.path.startswith('http'):
            credentials = repository.get_credentials()

            self.client = HgWebClient(
                path=repository.path,
                username=credentials['username'],
                password=credentials['password'],
                local_site=repository.local_site,
            )
        else:
            if not is_exe_in_path('hg'):
                # This is technically not the right kind of error, but it's the
                # pattern we use with all the other tools.
                raise ImportError

            if repository.local_site:
                local_site_name = repository.local_site.name
            else:
                local_site_name = None

            self.client = HgClient(
                path=repository.path,
                local_site_name=local_site_name,
                local_site=repository.local_site,
            )

    def get_file(
        self,
        path: str,
        revision: RevisionID = HEAD,
        context: Optional[FileLookupContext] = None,
        **kwargs,
    ) -> bytes:
        """Return the contents of a file from a repository.

        Args:
            path (str):
                The path to the file in the repository.

            revision (reviewboard.scmtools.core.Revision, optional):
                The revision to fetch. Subclasses should default this to
                :py:data:`HEAD`.

            context (reviewboard.scmtools.core.FileLookupContext, optional):
                Extra context used to help look up this file.

                This contains information about the HTTP request, requesting
                user, and parsed diff information, which may be useful as
                part of the repository lookup process.

            **kwargs (dict, unused):
                Additional keyword arguments. This is not currently used, but
                is available for future expansion.

        Returns:
            bytes:
            The returned file contents.

        Raises:
            reviewboard.scmtools.errors.FileNotFoundError:
                The file could not be found in the repository.

            reviewboard.scmtools.errors.InvalidRevisionFormatError:
                The ``revision`` or ``base_commit_id`` arguments were in an
                invalid format.
        """
        base_commit_id = context and context.base_commit_id

        return self.client.cat_file(
            path,
            str(revision),
            base_commit_id=base_commit_id)

    def parse_diff_revision(
        self,
        filename: bytes,
        revision: Union[Revision, bytes],
        *args,
        **kwargs,
    ) -> tuple[bytes, Union[Revision, bytes]]:
        """Parse and return a filename and revision from a diff.

        Args:
            filename (bytes):
                The filename as represented in the diff.

            revision (bytes):
                The revision as represented in the diff.

            *args (tuple, unused):
                Unused positional arguments.

            **kwargs (dict, unused):
                Unused keyword arguments.

        Returns:
            tuple:
            A tuple containing two items:

            1. The normalized filename as a byte string.
            2. The normalized revision as a byte string or a
               :py:class:`~reviewboard.scmtools.core.Revision`.
        """
        assert isinstance(filename, bytes), (
            'filename must be a byte string, not %s' % type(filename))
        assert isinstance(revision, bytes), (
            'revision must be a byte string, not %s' % type(revision))

        if filename == b'/dev/null':
            revision = PRE_CREATION

        return filename, revision or UNKNOWN

    def get_branches(self) -> list[Branch]:
        """Return open/inactive branches from repository.

        Returns:
            list of reviewboard.scmtools.core.Branch:
            The list of the branches.
        """
        return self.client.get_branches()

    def get_commits(
        self,
        branch: Optional[str] = None,
        start: Optional[str] = None,
    ) -> list[Commit]:
        """Return changesets from repository.

        Args:
            branch (str, optional):
                An identifier name of branch.

            start (str, optional):
                An optional changeset revision to start with.

        Returns:
            list of reviewboard.scmtools.core.Commit:
            The list of commits.
        """
        return self.client.get_commits(branch, start)

    def get_change(
        self,
        revision: str,
    ) -> Commit:
        """Return detailed information about a changeset.

        Receive changeset data and patch from repository.

        Args:
            revision (str):
                An identifier of changeset.

        Returns:
            reviewboard.scmtools.core.Commit:
            The commit object.
        """
        return self.client.get_change(revision)

    def get_parser(
        self,
        data: bytes,
    ) -> DiffParser:
        """Return a diff parser used to parse diff data.

        This will return either a :py:class:`HgDiffParser` or
        :py:class:`HgGitDiffParser` depending on the format of the diff data.

        Args:
            data (bytes):
                The diff data to parse.

        Returns:
            reviewboard.diffviewer.diffparser.BaseDiffParser:
            The diff parser used to parse this data.
        """
        diff_parser_cls = self._get_diff_parser_cls(data)

        return diff_parser_cls(data)

    @classmethod
    def date_tuple_to_iso8601(
        cls,
        data: tuple[int, int],
    ) -> str:
        """Return isoformat date from JSON tuple date.

        Args:
            data (tuple of int):
                 A 2-tuple, where the first item is a unix timestamp
                 and the second is the timezone offset.

        Returns:
            str:
            Date of given data in ISO 8601 format.
        """
        return force_str(datetime.utcfromtimestamp(
            data[0] + (data[1] * -1)).isoformat())

    @classmethod
    def check_repository(
        cls,
        path: str,
        username: (str | None) = None,
        password: (str | None) = None,
        local_site_name: (str | None) = None,
        local_site: (LocalSite | None) = None,
        **kwargs,
    ) -> None:
        """Check a repository configuration for validity.

        This should check if a repository exists and can be connected to.
        This will also check if the repository requires an HTTPS certificate.

        A failed result is returned as an exception. The exception may contain
        extra information, such as a human-readable description of the problem.
        If the repository is valid and can be connected to, no exception will
        be thrown.

        Args:
            path (str):
                The repository path.

            username (str, optional):
                The optional username for the repository.

            password (str, optional):
                The optional password for the repository.

            local_site_name (str, optional):
                The name of the :term:`Local Site` that owns this repository.
                This is optional.

            local_site (reviewboard.site.models.LocalSite, optional):
                The :term:`Local Site` instance that owns this repository.
                This is optional.

            **kwargs (dict, unused):
                Additional settings for the repository.

        Raises:
            reviewboard.scmtools.errors.AuthenticationError:
                The provided username/password or the configured SSH key could
                not be used to authenticate with the repository.

            reviewboard.scmtools.errors.RepositoryNotFoundError:
                A repository could not be found at the given path.

            reviewboard.scmtools.errors.SCMError:
                There was a generic error with the repository or its
                configuration.  Details will be provided in the error message.

            reviewboard.ssh.errors.BadHostKeyError:
                An SSH path was provided, but the host key for the repository
                did not match the expected key.

            reviewboard.ssh.errors.SSHError:
                An SSH path was provided, but there was an error establishing
                the SSH connection.

            reviewboard.ssh.errors.SSHInvalidPortError:
                An SSH path was provided, but the port specified was not a
                valid number.

            Exception:
                An unexpected exception has occurred. Callers should check
                for this and handle it.
        """
        result = urlparse(path)

        if result.scheme == 'ssh':
            raise SCMError('Mercurial over SSH is not supported.')

        super().check_repository(
            path=path,
            username=username,
            password=password,
            local_site_name=local_site_name,
            local_site=local_site,
            **kwargs,
        )

        # Create a client. This will fail if the repository doesn't exist.
        if result.scheme in ('http', 'https'):
            HgWebClient(path=path,
                        username=username,
                        password=password,
                        local_site=local_site)
        else:
            HgClient(path=path,
                     local_site_name=local_site_name,
                     local_site=local_site)

    def normalize_patch(
        self,
        patch: bytes,
        filename: str,
        revision: str,
    ) -> bytes:
        """Normalize the provided patch file.

        For Git-style diffs, this will update modes on new, changed, and
        deleted symlinks, stripping the symlink mode and making them appear
        as normal files. This will avoid any issues with applying the diff,
        and allow us to instead parse the symlink change as a regular file.

        Version Added:
            4.0.6

        Args:
            patch (bytes):
                The diff/patch file to normalize.

            filename (str):
                The name of the file being changed in the diff.

            revision (str):
                The revision of the file being changed in the diff.

        Returns:
            bytes:
            The resulting diff/patch file.
        """
        diff_parser_cls = self._get_diff_parser_cls(patch)

        if diff_parser_cls is HgGitDiffParser:
            patch = strip_git_symlink_mode(patch)

        return patch

    def _get_diff_parser_cls(
        self,
        data: bytes,
    ) -> type[DiffParser]:
        """Return the diff parser class used for this file.

        Version Added:
            4.0.6

        Args:
            data (bytes):
                The diff content.

        Returns:
            type:
            The diff parser used to parse the file.
        """
        hg_position = data.find(b'diff -r')
        git_position = data.find(b'diff --git')

        if git_position > -1 and (git_position < hg_position or
                                  hg_position == -1):
            return HgGitDiffParser
        else:
            return HgDiffParser


class HgDiffParser(DiffParser):
    """Diff parser for native Mercurial diffs."""

    # Plain Mercurial diffs don't include per-file revision information,
    # so we need to opt out of strong validation for diffs, and opt in to
    # heuristics that try to loosely validate a diff in a series of commits.
    # This also means that source revisions for files will be rewritten
    # after validation based on the most-likely commit.
    #
    # It's possible for merge commits or ill-formed commit chains to validate
    # but later fail to patch.
    has_per_file_revisions = False

    def __init__(
        self,
        data: bytes,
        **kwargs,
    ) -> None:
        """Initialize the parser.

        Args:
            data (bytes):
                The diff content to parse.

            **kwargs (dict):
                Keyword arguments to pass to the parent class.

        Raises:
            TypeError:
                The provided ``data`` argument was not a ``bytes`` type.
        """
        super().__init__(data, uses_commit_ids_as_revisions=True)

    def parse_special_header(
        self,
        linenum: int,
        parsed_file: ParsedDiffFile,
    ) -> int:
        """Parse a special diff header marking the start of a new file's info.

        This looks for some special markers found in Mercurial diffs, trying
        to find a ``Parent`` or a ``diff -r`` line.

        A ``Parent`` line specifies a changeset ID that will be used as the
        source revision for all files.

        A ``diff -r`` line contains information identifying the file's name
        and other details.

        Args:
            linenum (int):
                The line number to begin parsing.

            parsed_file (reviewboard.diffviewer.parser.ParsedDiffFile):
                The file currently being parsed.

        Returns:
            int:
            The next line number to parse.

        Raises:
            reviewboard.diffviewer.errors.DiffParserError:
                There was an error parsing the special header. This may be
                a corrupted diff, or an error in the parsing implementation.
                Details are in the error message.
        """
        diff_line = self.lines[linenum]
        split_line = diff_line.split()

        if diff_line.startswith(b'# Parent') and len(split_line) == 3:
            self.parsed_diff_change.parent_commit_id = split_line[2]
        elif diff_line.startswith(b'diff -r'):
            # A diff between two revisions are in the following form:
            #
            #     diff -r abcdef123456 -r 123456abcdef filename
            #
            # A diff between a revision and the working copy:
            #
            #     diff -r abcdef123456 filename
            try:
                # Ordinary hg diffs don't record renames, so a new file
                # is always equivalent to an old file.
                if len(split_line) > 4 and split_line[3] == b'-r':
                    # Committed revision
                    name_start_ix = 5
                    parsed_file.modified_file_details = split_line[4]
                else:
                    # Uncommitted revision
                    name_start_ix = 3
                    parsed_file.modified_file_details = b'Uncommitted'

                filename = b' '.join(split_line[name_start_ix:])

                parsed_file.orig_filename = filename
                parsed_file.orig_file_details = split_line[2]
                parsed_file.modified_filename = filename

                self.parsed_diff_change.parent_commit_id = split_line[2]
            except ValueError:
                raise DiffParserError('The diff file is missing revision '
                                      'information',
                                      linenum=linenum)

            linenum += 1

        return linenum

    def parse_diff_header(
        self,
        linenum: int,
        parsed_file: ParsedDiffFile,
    ) -> int:
        """Parse a standard header before changes made to a file.

        This will look for information indicating if the file is a binary file,
        and then attempt to parse the standard diff headers (``---`` and
        ``+++`` lines).

        Args:
            linenum (int):
                The line number to begin parsing.

            parsed_file (reviewboard.diffviewer.parser.ParsedDiffFile):
                The file currently being parsed.

        Returns:
            int:
            The next line number to parse.
        """
        lines = self.lines

        if (linenum <= len(lines) and
            lines[linenum].startswith(b'Binary file ')):
            parsed_file.binary = True
            linenum += 1

        if (linenum + 1 < len(lines) and
            (lines[linenum].startswith(b'--- ') and
             lines[linenum + 1].startswith(b'+++ '))):
            # Check if this is a new file.
            if lines[linenum].split()[1] == b'/dev/null':
                parsed_file.orig_file_details = PRE_CREATION

            # Check if this is a deleted file.
            if lines[linenum + 1].split()[1] == b'/dev/null':
                parsed_file.deleted = True

            linenum += 2

        return linenum


class HgGitDiffParser(GitDiffParser):
    """Diff Parser for Git diffs generated by Mercurial.

    Mercurial-generated Git diffs contain additional metadata that need to
    be parsed in order to properly locate changes to files in a repository.
    """

    # Git-style Mercurial diffs don't include per-file revision information,
    # so we need to opt out of strong validation for diffs.
    #
    # See HgDiffParser for more details on the implications here.
    has_per_file_revisions = False

    def __init__(
        self,
        data: bytes,
        **kwargs,
    ) -> None:
        """Initialize the parser.

        Args:
            data (bytes):
                The diff content to parse.

            **kwargs (dict):
                Keyword arguments to pass to the parent class.

        Raises:
            TypeError:
                The provided ``data`` argument was not a ``bytes`` type.
        """
        super().__init__(data, uses_commit_ids_as_revisions=True)

    def parse(self) -> list[ParsedDiffFile]:
        """Parse the diff.

        This will parse the diff, looking for changes to the file.

        It special-cases the default Git diff parsing to check for any
        ``# Node ID`` or ``# Parent`` lines found at the beginning of the
        file, which specify the new commit ID and the base commit ID,
        respectively.

        Returns:
            list of ParsedDiffFile:
            The resulting list of files.

        Raises:
            reviewboard.diffviewer.errors.DiffParserError:
                There was an error parsing part of the diff. This may be a
                corrupted diff, or an error in the parsing implementation.
                Details are in the error message.
        """
        lines = self.lines
        linenum = 0

        while lines[linenum].startswith(b'#'):
            line = lines[linenum]
            split_line = line.split()
            linenum += 1

            if line.startswith(b'# Node ID') and len(split_line) == 4:
                self.new_commit_id = split_line[3]
            elif line.startswith(b'# Parent') and len(split_line) == 3:
                self.base_commit_id = split_line[2]
                self.parsed_diff_change.parent_commit_id = split_line[2]

        diff_files = super().parse()

        # hg diff --git generally doesn't include index lines for files, and so
        # by default GitDiffParser will use ``self.base_commit_id`` and
        # ``self.new_commit_id`` for individual file revisions. The exception
        # is for binary files, which get an index line containing the binary
        # file's SHA. This is meaningless for Mercurial. Go back through and
        # overwrite those with the revision IDs.
        for f in diff_files:
            if f.binary:
                if f.orig_file_details != PRE_CREATION:
                    f.orig_file_details = self.base_commit_id

                f.modified_file_details = self.new_commit_id

        return diff_files


class HgWebClient(SCMClient):
    """Client for communicating with HgWeb."""

    FULL_FILE_URL = '%(url)s/%(rawpath)s/%(revision)s/%(quoted_path)s'

    def __init__(
        self,
        *args,
        **kwargs,
    ) -> None:
        """Initialize the client.

        Args:
            *args (tuple):
                Positional arguments to pass to the parent method.

            **kwargs (dict):
                Keyword arguments to pass to the parent method.
        """
        super().__init__(*args, **kwargs)

        path = self.path
        self.path_stripped = path.rstrip('/')

        logger.debug('Initialized HgWebClient with url=%r, username=%r',
                     path, self.username)

    def cat_file(
        self,
        path: str,
        rev: RevisionID = 'tip',
        base_commit_id: Optional[str] = None,
    ) -> bytes:
        """Return the content of a file.

        Args:
            path (str):
                The path of the file to return.

            rev (reviewboard.scmtools.core.RevisionID, optional):
                The revision of the file to return.

            base_commit_id (str, optional):
                The base commit, if using.

        Returns:
            bytes:
            The content of the file at the given revision.

        Raises:
            reviewboard.scmtools.errors.FileNotFoundError:
                The file was not found at the given name and revision.
        """
        # If the base commit id is provided it should override anything
        # that was parsed from the diffs.
        if rev != PRE_CREATION and base_commit_id is not None:
            rev = base_commit_id

        if rev == HEAD or rev == UNKNOWN:
            rev = "tip"
        elif rev == PRE_CREATION:
            rev = ""

        for rawpath in ["raw-file", "raw", "hg-history"]:
            try:
                url = self.FULL_FILE_URL % {
                    'url': self.path_stripped,
                    'rawpath': rawpath,
                    'revision': rev,
                    'quoted_path': urllib_quote(path.lstrip('/')),
                }

                data = self.get_file_http(url, path, rev)

                if data is not None:
                    return data
            except Exception:
                # It failed. Error was logged and we may try again.
                pass

        raise FileNotFoundError(path, rev)

    def get_branches(self) -> list[Branch]:
        """Return open/inactive branches from hgweb in JSON.

        Returns:
            list of reviewboard.scmtools.core.Branch:
            A list of the branches.
        """
        results = []

        try:
            rsp = self._get_http_json('%s/json-branches' % self.path_stripped)
        except Exception as e:
            logger.exception('Cannot load branches from hgweb: %s', e)
            return results

        if rsp:
            results = [
                Branch(
                    id=data['branch'],
                    commit=data['node'],
                    default=(data['branch'] == 'default'))
                for data in rsp['branches']
                if data['status'] != 'closed'
            ]

        return results

    def _get_commit(
        self,
        revision: str,
    ) -> Optional[Commit]:
        """Return detailed information about a single changeset.

        Receive changeset from hgweb in JSON format.

        Args:
            revision (str):
                An identifier of changeset.

        Returns:
            reviewboard.scmtools.core.Commit:
            The commit object.
        """
        try:
            rsp = self._get_http_json(
                f'{self.path_stripped}/json-rev/{revision}')
        except Exception as e:
            logger.exception('Cannot load detail of changeset from hgweb: %s',
                             e)
            return None

        if not rsp:
            return None

        try:
            parent = rsp['parents'][0]

            if parent == INITIAL_COMMIT_ID:
                parent = ''
        except IndexError:
            parent = ''

        return Commit(id=rsp['node'],
                      message=rsp['desc'],
                      author_name=rsp['user'],
                      date=HgTool.date_tuple_to_iso8601(rsp['date']),
                      parent=parent)

    def get_commits(
        self,
        branch: Optional[str] = None,
        start: Optional[str] = None,
    ) -> list[Commit]:
        """Return detailed information about a changeset.

        Receive changeset from hgweb in JSON format.

        Args:
            branch (str, optional):
                An optional branch name to filter by.

            start (str, optional):
                An optional changeset revision to start with.

        Returns:
            list of reviewboard.scmtools.core.Commit:
            The list of commit objects.
        """
        query_parts = []

        if start:
            query_parts.append('ancestors(%s)' % start)

        query_parts.append('branch(%s)' % (branch or '.'))

        query = '+and+'.join(query_parts)

        try:
            rsp = self._get_http_json(
                f'{self.path_stripped}/json-log/?rev={query}')
        except Exception as e:
            logger.exception('Cannot load commits from hgweb: %s', e)
            return []

        results = []

        if rsp:
            for data in rsp['entries']:
                try:
                    parent = data['parents'][0]

                    if parent == INITIAL_COMMIT_ID:
                        parent = ''
                except IndexError:
                    parent = ''

                iso8601 = HgTool.date_tuple_to_iso8601(data['date'])
                changeset = Commit(id=data['node'],
                                   message=data['desc'],
                                   author_name=data['user'],
                                   date=iso8601,
                                   parent=parent)
                results.append(changeset)

        return results

    def get_change(
        self,
        revision: str,
    ) -> Commit:
        """Return detailed information about a changeset.

        This method retrieves the patch in JSON format from hgweb.

        Args:
            revision (str):
                An identifier of changeset

        Returns:
            reviewboard.scmtools.core.Commit:
            The commit object.
        """
        try:
            contents = self.get_file_http(
                url=f'{self.path_stripped}/raw-rev/{revision}',
                path='',
                revision='')
        except Exception as e:
            logger.exception('Cannot load patch from hgweb: %s', e)
            raise SCMError('Cannot load patch from hgweb')

        if contents:
            changeset = self._get_commit(revision)

            if changeset:
                changeset.diff = contents
                return changeset

        logger.error('Cannot load changeset %s from hgweb', revision)
        raise SCMError('Cannot load changeset %s from hgweb' % revision)

    def _get_http_json(
        self,
        url: str,
    ) -> Optional[dict[str, Any]]:
        """Return a JSON response from an HgWeb API endpoint.

        Args:
            url (str):
                The URL of the JSON payload to fetch.

        Returns:
            object:
            The deserialized JSON data.

        Raises:
            Exception:
                Fetching the file, decoding the payload, or deserializing
                the JSON has failed.
        """
        contents = self.get_file_http(url=url,
                                      path='',
                                      revision='',
                                      mime_type='application/json')

        if contents is None or contents == b'not yet implemented':
            return None

        return json.loads(contents.decode('utf-8'))


class HgClient(SCMClient):
    """Client implementation for using the hg tool."""

    COMMITS_PAGE_LIMIT = '31'

    def __init__(
        self,
        path: str,
        local_site_name: str | None,
        **kwargs,
    ) -> None:
        """Initialize the client.

        Version Changed:
            8.0:
            Added the ``**kwargs`` argument.

        Args:
            path (str):
                The path to the repository.

            local_site_name (str, optional):
                The name of the Local Site that the repository is part of.

            **kwargs (dict):
                Additional keyword arguments for the parent method.

                Version Added:
                    8.0
        """
        super().__init__(path, **kwargs)

        self.default_args = None
        self.local_site_name = local_site_name

    def cat_file(
        self,
        path: str,
        rev: RevisionID = 'tip',
        base_commit_id: Optional[str] = None,
    ) -> bytes:
        """Return the content of a file.

        Args:
            path (str):
                The path of the file to return.

            rev (reviewboard.scmtools.core.RevisionID, optional):
                The revision of the file to return.

            base_commit_id (str, optional):
                The base commit, if using.

        Returns:
            bytes:
            The content of the file at the given revision.

        Raises:
            reviewboard.scmtools.errors.FileNotFoundError:
                The file was not found at the given name and revision.
        """
        # If the base commit id is provided it should override anything
        # that was parsed from the diffs.
        if rev != PRE_CREATION and base_commit_id is not None:
            rev = base_commit_id

        if rev == HEAD:
            rev = "tip"
        elif rev == PRE_CREATION:
            rev = ""

        if path:
            with self._run_hg(['cat', '--rev', str(rev), path]) as p:
                assert p.stdout is not None
                contents = p.stdout.read()
                failure = p.wait()

            if not failure:
                return contents

        raise FileNotFoundError(path, rev)

    def get_branches(self) -> list[Branch]:
        """Return open/inactive branches from repository in JSON.

        Returns:
            list of reviewboard.scmtools.core.Branch:
            The list of the branches.
        """
        with self._run_hg(['branches', '--template', 'json']) as p:
            if p.wait() != 0:
                assert p.stderr is not None
                raise SCMError('Cannot load branches: %s' % p.stderr.read())

            assert p.stdout is not None

            return [
                Branch(
                    id=data['branch'],
                    commit=data['node'],
                    default=(data['branch'] == 'default'))
                for data in json.loads(force_str(p.stdout.read()))
                if not data['closed']
            ]

    def _get_commits(
        self,
        revset: list[str],
    ) -> list[Commit]:
        """Return a list of commit objects.

        This method calls the given revset and parses the returned
        JSON data to retrieve detailed information about changesets.

        Args:
            revset (list of str):
                Hg command line that will be executed with JSON
                template as log command.

        Returns:
            list of reviewboard.scmtools.core.Commit:
            The list of commit objects.
        """
        cmd = ['log'] + revset + ['--template', 'json']

        with self._run_hg(cmd) as p:
            if p.wait() != 0:
                assert p.stderr is not None
                raise SCMError('Cannot load commits: %s' % p.stderr.read())

            results = []

            assert p.stdout is not None

            for data in json.loads(force_str(p.stdout.read())):
                try:
                    parent: Optional[str] = force_str(data['parents'][0])

                    if parent == INITIAL_COMMIT_ID:
                        parent = ''
                except IndexError:
                    parent = None

                results.append(Commit(
                    id=data['node'],
                    message=data['desc'],
                    author_name=data['user'],
                    date=HgTool.date_tuple_to_iso8601(data['date']),
                    parent=parent))

        return results

    def get_commits(
        self,
        branch: Optional[str] = None,
        start: Optional[str] = None,
    ) -> list[Commit]:
        """Return changesets from repository in JSON.

        Args:
            branch (str, optional):
                An identifier name of branch.

            start (str, optional):
                An optional changeset revision to start with.

        Returns:
            list of reviewboard.scmtools.core.Commit:
            The list of commit objects.
        """
        revisions = ''

        if start:
            revisions = '-r%s:0' % start

        revset = [revisions, '-l', self.COMMITS_PAGE_LIMIT]

        if branch:
            revset.extend(('-b', branch))

        return self._get_commits(revset)

    def get_change(
        self,
        revision: str,
    ) -> Commit:
        """Return detailed information about a changeset.

        Receive changeset data and patch from repository in JSON.

        Args:
            revision (str):
                An identifier of changeset.

        Returns:
            reviewboard.scmtools.core.Commit:
            The commit object.
        """
        revset = ['-r', revision]
        changesets = self._get_commits(revset)

        if changesets:
            commit = changesets[0]
            cmd = ['diff', '-c', revision]

            with self._run_hg(cmd) as p:

                if p.wait() != 0:
                    assert p.stderr is not None

                    e = p.stderr.read()
                    raise SCMError(f'Cannot load patch {revision}: {e}')

                assert p.stdout is not None

                commit.diff = p.stdout.read()

            return commit

        raise SCMError('Cannot load changeset %s' % revision)

    def _calculate_default_args(self) -> None:
        self.default_args = [
            '--noninteractive',
            '--repository', self.path,
            '--cwd', self.path,
        ]

        # We need to query hg for the current SSH configuration. Note
        # that _run_hg is calling this function, and this function is then
        # (through _get_hg_config) calling _run_hg, but it's okay. Due to
        # having set a good default for self.default_args above, there's no
        # issue of an infinite loop.
        hg_ssh = self._get_hg_config('ui.ssh')

        if not hg_ssh:
            logger.debug('Using rbssh for Mercurial')

            if self.local_site_name:
                hg_ssh = 'rbssh --rb-local-site=%s' % self.local_site_name
            else:
                hg_ssh = 'rbssh'

            self.default_args.extend([
                '--config', 'ui.ssh=%s' % hg_ssh,
            ])
        else:
            logger.debug('Found configured ssh for Mercurial: %s', hg_ssh)

    def _get_hg_config(
        self,
        config_name: str,
    ) -> Optional[str]:
        with self._run_hg(['showconfig', config_name]) as p:
            assert p.stdout is not None

            contents = p.stdout.read()
            failure = p.wait()

        if failure:
            # Just assume it's empty.
            return None

        return contents.strip()

    def _run_hg(
        self,
        args: list[str],
    ) -> subprocess.Popen:
        """Run the Mercurial command, returning a subprocess.Popen.

        Args:
            args (list of str):
                The arguments to add to the :command:`hg` command.

        Returns:
            subprocess.Popen:
            The Popen object.
        """
        if not self.default_args:
            self._calculate_default_args()

        assert self.default_args is not None

        return SCMTool.popen(
            ['hg'] + self.default_args + args,
            local_site_name=self.local_site_name)



"""Build backend for Review Board extension packages.

Version Added:
    8.0
"""

from __future__ import annotations

import os
import sys
from contextlib import contextmanager
from typing import Iterator, Optional

from setuptools import build_meta


@contextmanager
def _prepare_package_files() -> Iterator[None]:
    """Prepare files for the package.

    This will create a :file:`setup.py` if one doesn't already exist and
    ensure the extension's module can be found in the search path.

    Context:
        The package files and environment will be ready for building.
    """
    setup_py_created: bool = False

    if not os.path.exists('setup.py'):
        # There's no setup.py here, so we'll want to create one that calls
        # our version.
        #
        # This is the same approach that setuptools (at least as of v74)
        # uses. The build backend is still a wrapper around calling setup().
        setup_py_created = True

        with open('setup.py', 'w') as fp:
            fp.write(
                'from reviewboard.extensions.packaging.setuptools_backend'
                ' import setup\n'
                '\n'
                'setup()\n'
            )

    # Make sure the extension's module can be found in this directory.
    sys.path.insert(0, os.getcwd())

    try:
        yield
    finally:
        if setup_py_created:
            os.unlink('setup.py')

        sys.path.pop(0)


def build_sdist(
    sdist_directory: str,
    config_settings: Optional[dict] = None,
) -> str:
    """Build a source distribution.

    This will prepare a setup.py to run our extension-building logic, and
    then build the distribution.

    Args:
        sdist_directory (str):
            The directory where the source distribution will be placed.

        config_settings (dict, optional):
            Configuration settings to pass to Setuptools.

    Returns:
        str:
        The basename for the generated source distribution file.
    """
    with _prepare_package_files():
        return build_meta.build_sdist(sdist_directory,
                                      config_settings)


def build_wheel(
    wheel_directory: str,
    config_settings: Optional[dict] = None,
    metadata_directory: Optional[str] = None,
) -> str:
    """Build a wheel.

    This will prepare a setup.py to run our extension-building logic, and
    then build the distribution.

    Args:
        wheel_directory (str):
            The directory where the wheel will be placed.

        config_settings (dict, optional):
            Configuration settings to pass to Setuptools.

    Returns:
        str:
        The basename for the generated wheel file.
    """
    with _prepare_package_files():
        return build_meta.build_wheel(wheel_directory,
                                      config_settings,
                                      metadata_directory)


get_requires_for_build_editable = build_meta.get_requires_for_build_editable
get_requires_for_build_sdist = build_meta.get_requires_for_build_sdist
get_requires_for_build_wheel = build_meta.get_requires_for_build_wheel
build_editable = build_meta.build_editable
prepare_metadata_for_build_editable = \
    build_meta.prepare_metadata_for_build_editable
prepare_metadata_for_build_wheel = build_meta.prepare_metadata_for_build_wheel



"""Base patterns and support for defining expected queries for unit tests.

Version Added:
    5.0.7
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from typing_extensions import TypedDict

if TYPE_CHECKING:
    from django.db.models import Q
    from django_assert_queries.query_comparator import ExpectedQueries
    from typing_extensions import NotRequired


class ExpectedQResult(TypedDict):
    """An expected Q-expression result from a utility function.

    This should be used by unit test utility functions that build expected Q
    expressions in order to provide consistent results with typing.

    Version Added:
        5.0.7
    """

    #: The expected Q expression.
    q: Q

    #: The set of tables involved in the Q expression.
    tables: set[str]

    #: The types of any joined tables.
    join_types: NotRequired[dict[str, str]]

    #: Any queries that should be expected before a query using this Q.
    prep_equeries: NotRequired[ExpectedQueries]

    #: Any subqueries that are generated by this Q-expression.
    subqueries: NotRequired[ExpectedQueries]



"""Hosting service for Beanstalk."""

from __future__ import annotations

import json
import logging
import os
from collections import defaultdict
from urllib.error import HTTPError, URLError
from urllib.parse import quote

from django import forms
from django.http import HttpResponse
from django.urls import path
from django.utils.translation import gettext_lazy as _
from django.views.decorators.http import require_POST
from housekeeping import deprecate_non_keyword_only_args

from reviewboard.admin.server import get_server_url
from reviewboard.deprecation import RemovedInReviewBoard10_0Warning
from reviewboard.hostingsvcs.base.forms import (
    BaseHostingServiceAuthForm,
    BaseHostingServiceRepositoryForm,
)
from reviewboard.hostingsvcs.base.hosting_service import BaseHostingService
from reviewboard.hostingsvcs.hook_utils import (close_all_review_requests,
                                                get_review_request_id)
from reviewboard.scmtools.crypto_utils import (decrypt_password,
                                               encrypt_password)
from reviewboard.scmtools.errors import FileNotFoundError
from reviewboard.scmtools.svn.utils import (collapse_svn_keywords,
                                            has_expanded_svn_keywords)


logger = logging.getLogger(__name__)


class BeanstalkAuthForm(BaseHostingServiceAuthForm):
    """Authentication form for the Beanstalk hosting service.

    This replaces some of the help text to make setup a bit easier.
    """

    class Meta:
        labels = {
            'hosting_account_password': 'Access token',
        }

        help_texts = {
            'hosting_account_username': _(
                'Your Beanstalk username. This is case-sensitive and will '
                'not be your Beanstalk e-mail address.'
            ),
            'hosting_account_password': _(
                'A pre-generated access token used to log into your account '
                'via an API. You can generate these on Beanstalk by clicking '
                'your name in the top-right of any page, clicking "Access '
                'Tokens," and then clicking "Generate a token."'
            ),
        }


class BeanstalkForm(BaseHostingServiceRepositoryForm):
    beanstalk_account_domain = forms.CharField(
        label=_('Beanstalk account domain'),
        max_length=64,
        required=True,
        widget=forms.TextInput(attrs={'size': '60'}),
        help_text=_('This is the <tt>domain</tt> part of '
                    '<tt>domain.beanstalkapp.com</tt>'))

    beanstalk_repo_name = forms.CharField(
        label=_('Repository name'),
        max_length=64,
        required=True,
        widget=forms.TextInput(attrs={'size': '60'}))


class BeanstalkHookViews(object):
    """Container class for hook views."""

    @staticmethod
    @require_POST
    def process_post_receive_hook(request, local_site_name=None,
                                  repository_id=None,
                                  hosting_service_id=None,
                                  *args, **kwargs):
        """Close review requests as submitted automatically after a push.

        Args:
            request (django.http.HttpRequest):
                The request from the Beanstalk webhook.

            local_site_name (unicode, optional):
                The local site name, if available.

            repository_id (int, optional):
                The pk of the repository, if available.

            hosting_service_id (unicode, optional):
                The name of the hosting service.

            *args (list):
                Additional postitional arguments.

            **kwargs (dict):
                Additional keyword arguments.

        Returns:
            django.http.HttpResponse:
            The HTTP response.
        """
        try:
            server_url = get_server_url(request=request)

            # Check if it's a git or an SVN repository and close accordingly.
            if 'payload' in request.POST:
                payload = json.loads(request.POST['payload'])
                BeanstalkHookViews._close_git_review_requests(
                    payload, server_url, local_site_name, repository_id,
                    hosting_service_id)
            else:
                payload = json.loads(request.POST['commit'])
                BeanstalkHookViews._close_svn_review_request(
                    payload, server_url, local_site_name, repository_id,
                    hosting_service_id)
        except KeyError as e:
            logger.error('There is no JSON payload in the POST request.: %s',
                         e,
                         extra={'request': request})
            return HttpResponse(status=415)
        except ValueError as e:
            logger.error('The payload is not in JSON format: %s', e,
                         extra={'request': request})
            return HttpResponse(status=415)

        return HttpResponse()

    @staticmethod
    def _close_git_review_requests(payload,
                                   server_url,
                                   local_site_name,
                                   repository_id,
                                   hosting_service_id):
        """Close all review requests for the git repository.

        A git payload may contain multiple commits. If a commit's commit
        message does not contain a review request ID, it closes based on
        it's commit id.

        Args:
            payload (dict):
                The decoded webhook payload.

            server_url (unicode):
                The current server URL.

            local_site_name (unicode):
                The local site name, if available.

            repository_id (int):
                The pk of the repository, if available.

            hosting_service_id (unicode):
                The name of the hosting service.
        """
        review_id_to_commits_map = defaultdict(list)
        branch_name = payload.get('branch')

        if not branch_name:
            return review_id_to_commits_map

        commits = payload.get('commits', [])

        for commit in commits:
            commit_hash = commit.get('id')
            commit_message = commit.get('message')
            review_request_id = get_review_request_id(
                commit_message, server_url, commit_hash)
            commit_entry = '%s (%s)' % (branch_name, commit_hash[:7])
            review_id_to_commits_map[review_request_id].append(commit_entry)

        close_all_review_requests(review_id_to_commits_map, local_site_name,
                                  repository_id, hosting_service_id)

    @staticmethod
    def _close_svn_review_request(payload,
                                  server_url,
                                  local_site_name,
                                  repository_id,
                                  hosting_service_id):
        """Close the review request for an SVN repository.

        The SVN payload contains one commit. If the commit's message does not
        contain a review request ID, this will not close any review requests.

        Args:
            payload (dict):
                The decoded webhook payload.

            server_url (unicode):
                The current server URL.

            local_site_name (unicode):
                The local site name, if available.

            repository_id (int):
                The pk of the repository, if available.

            hosting_service_id (unicode):
                The name of the hosting service.
        """
        review_id_to_commits_map = defaultdict(list)
        commit_message = payload.get('message')
        branch_name = payload.get('changeset_url', 'SVN Repository')
        revision = '%s %d' % ('Revision: ', payload.get('revision'))
        review_request_id = get_review_request_id(commit_message, server_url,
                                                  None)
        commit_entry = '%s (%s)' % (branch_name, revision)
        review_id_to_commits_map[review_request_id].append(commit_entry)
        close_all_review_requests(review_id_to_commits_map, local_site_name,
                                  repository_id, hosting_service_id)


class Beanstalk(BaseHostingService):
    """Hosting service support for Beanstalk.

    Beanstalk is a source hosting service that supports Git and Subversion
    repositories. It's available at http://beanstalkapp.com/.
    """
    name = 'Beanstalk'
    hosting_service_id = 'beanstalk'

    needs_authorization = True
    supports_bug_trackers = False
    supports_repositories = True
    supported_scmtools = ['Git', 'Subversion']

    form = BeanstalkForm
    auth_form = BeanstalkAuthForm

    repository_fields = {
        'Git': {
            'path': 'git@%(beanstalk_account_domain)s'
                    '.beanstalkapp.com:/%(beanstalk_account_domain)s/'
                    '%(beanstalk_repo_name)s.git',
            'mirror_path': 'https://%(beanstalk_account_domain)s'
                           '.git.beanstalkapp.com/%(beanstalk_repo_name)s.git',
        },
        'Subversion': {
            'path': 'https://%(beanstalk_account_domain)s'
                    '.svn.beanstalkapp.com/%(beanstalk_repo_name)s/',
        },
    }

    repository_url_patterns = [
        path('hooks/post-receive/',
             BeanstalkHookViews.process_post_receive_hook),
    ]

    @deprecate_non_keyword_only_args(RemovedInReviewBoard10_0Warning)
    def check_repository(
        self,
        *,
        beanstalk_account_domain: str,
        beanstalk_repo_name: str,
        **kwargs,
    ) -> None:
        """Check the validity of a repository.

        This will perform an API request against Beanstalk to get
        information on the repository. This will throw an exception if
        the repository was not found, and return cleanly if it was found.

        Version Changed:
            8.0:
            Made arguments keyword-only.

        Args:
            beanstalk_account_domain (str):
                The domain for the Beanstalk account.

            beanstalk_repo_name (str):
                The name of the repository.

            **kwargs (dict, unused):
                Additional information passed by the repository form.
        """
        self._api_get_repository(beanstalk_account_domain, beanstalk_repo_name)

    @deprecate_non_keyword_only_args(RemovedInReviewBoard10_0Warning)
    def authorize(
        self,
        *,
        username: str | None,
        password: str | None,
        **kwargs,
    ) -> None:
        """Authorize the Beanstalk repository.

        Beanstalk uses HTTP Basic Auth for the API, so this will store the
        provided password, encrypted, for use in later API requests.

        Version Changed:
            8.0:
            Made arguments keyword-only.

        Args:
            username (str):
                The username for the account.

            password (str):
                The password for the account.

            **kwargs (dict, unused):
                Extra keyword arguments containing values from the
                repository's configuration.
        """
        self.account.data['password'] = encrypt_password(password)
        self.account.save()

    def is_authorized(self):
        """Determines if the account has supported authorization tokens.

        This just checks if there's a password set on the account.
        """
        return self.account.data.get('password', None) is not None

    def get_password(self):
        """Returns the password for this account.

        This is needed for API calls and for Subversion.
        """
        return decrypt_password(self.account.data['password'])

    def get_file(self, repository, path, revision, base_commit_id=None,
                 *args, **kwargs):
        """Fetches a file from Beanstalk.

        This will perform an API request to fetch the contents of a file.

        If using Git, this will expect a base commit ID to be provided.
        """
        try:
            contents = self._api_get_node(repository, path, revision,
                                          base_commit_id, contents=True)
        except URLError:
            raise FileNotFoundError(path, revision)

        # On Subversion repositories, we may need to expand properties within
        # the file, like ``$Id$``. We only want to do this if we see keywords.
        if repository.tool.name == 'Subversion':
            contents = self._normalize_svn_file_content(
                repository, contents, path, revision)

        return contents

    def get_file_exists(self, repository, path, revision, base_commit_id=None,
                        *args, **kwargs):
        """Determines if a file exists.

        This will perform an API request to fetch the metadata for a file.

        If using Git, this will expect a base commit ID to be provided.
        """
        try:
            self._api_get_node(repository, path, revision, base_commit_id)

            return True
        except (HTTPError, URLError, FileNotFoundError):
            return False

    def normalize_patch(self, repository, patch, filename, revision):
        """Normalize a diff/patch file before it's applied.

        If working with a Subversion repository, then diffs being put up
        for review may have expanded keywords in them. This may occur if
        the file was diffed against a repository that did not (at that time)
        list those keywords in the ``svn:keywords`` property. We need to
        collapse these down.

        For non-Subversion repositories, the default behavior of the
        repository backend is used.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository the patch is meant to apply to.

            patch (bytes):
                The diff/patch file to normalize.

            filename (unicode):
                The name of the file being changed in the diff.

            revision (unicode):
                The revision of the file being changed in the diff.

        Returns:
            bytes:
            The resulting diff/patch file.
        """
        if repository.tool.name == 'Subversion':
            return self._normalize_svn_file_content(
                repository, patch, filename, revision)
        else:
            return super(Beanstalk, self).normalize_patch(
                repository, patch, filename, revision)

    def _normalize_svn_file_content(self, repository, contents, path,
                                    revision):
        """Post-process a file pertaining to a Subversion repository.

        This is common code that handles collapsing keywords for files fetched
        from or diffed against a Subversion repository.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository the content is for.

            contents (bytes):
                The file content to normalize.

            path (unicode):
                The path to the file.

            revision (unicode):
                The revision of the file.

        Returns:
            bytes:
            The resulting file.
        """
        if has_expanded_svn_keywords(contents):
            try:
                props = self._api_get_svn_props(repository, path, revision)
            except URLError:
                props = None

            if props and 'svn:keywords' in props:
                contents = collapse_svn_keywords(
                    contents,
                    props['svn:keywords'].encode('utf-8'))

        return contents

    def _api_get_repository(self, account_domain, repository_name):
        url = self._build_api_url(account_domain,
                                  'repositories/%s.json' % repository_name)

        return self._api_get(url)

    def _api_get_node(self, repository, path, revision, base_commit_id,
                      contents=False):
        # Unless we're fetching raw content, we optimistically want to
        # grab the metadata for the file. That's going to be a lot smaller
        # than the file contents in most cases. However, we can only do that
        # with a base_commit_id. If we don't have that, we fall back on
        # fetching the full file contents.
        is_git = (repository.tool.name == 'Git')

        if is_git and (contents or not base_commit_id):
            url_path = ('blob?id=%s&name=%s'
                        % (quote(revision), quote(os.path.basename(path))))
            raw_content = True
        else:
            if is_git:
                expected_revision = base_commit_id
            else:
                expected_revision = revision

            url_path = ('node.json?path=%s&revision=%s'
                        % (quote(path), quote(expected_revision)))

            if contents:
                url_path += '&contents=true'

            raw_content = False

        url = self._build_api_url(
            self._get_repository_account_domain(repository),
            'repositories/%s/%s'
            % (repository.extra_data['beanstalk_repo_name'], url_path))

        result = self._api_get(url, raw_content=raw_content)

        if not raw_content and contents:
            result = result['contents'].encode('utf-8')

        return result

    def _api_get_svn_props(self, repository, path, revision):
        """Return the SVN properties for a file in the repository.

        This will query for all SVN properties set for a particular file,
        returning them as a dictionary mapping property names to values.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The Subversion repository containing the file.

            path (unicode):
                The path to the file to retrieve properties for.

            revision (unicode):
                The revision of the file.

        Returns:
            dict:
            A mapping of property names to values.
        """
        url = self._build_api_url(
            self._get_repository_account_domain(repository),
            'repositories/%s/props.json?path=%s&revision=%s'
            % (repository.extra_data['beanstalk_repo_name'],
               quote(path), quote(revision)))

        result = self._api_get(url)

        return result.get('svn_properties', {})

    def _build_api_url(self, account_domain, url):
        return 'https://%s.beanstalkapp.com/api/%s' % (account_domain, url)

    def _get_repository_account_domain(self, repository):
        return repository.extra_data['beanstalk_account_domain']

    def _api_get(self, url, raw_content=False):
        try:
            response = self.client.http_get(
                url,
                username=self.account.username,
                password=self.get_password())

            if raw_content:
                return response.data
            else:
                return response.json
        except HTTPError as e:
            data = e.read()

            try:
                rsp = json.loads(data)
            except Exception:
                rsp = None

            if rsp and 'errors' in rsp:
                raise Exception('; '.join(rsp['errors']))
            else:
                raise Exception(str(e))



from django.core.exceptions import ObjectDoesNotExist
from django.db.models.query import Q
from djblets.util.decorators import augment_method_from
from djblets.webapi.decorators import (webapi_login_required,
                                       webapi_response_errors,
                                       webapi_request_fields)
from djblets.webapi.errors import (DOES_NOT_EXIST, INVALID_FORM_DATA,
                                   NOT_LOGGED_IN, PERMISSION_DENIED)
from djblets.webapi.fields import IntFieldType

from reviewboard.diffviewer.features import dvcs_feature
from reviewboard.diffviewer.models import FileDiff
from reviewboard.webapi.decorators import webapi_check_local_site
from reviewboard.webapi.resources import resources
from reviewboard.webapi.resources.base_diff_comment import \
    BaseDiffCommentResource


class ReviewDiffCommentResource(BaseDiffCommentResource):
    """Provides information on diff comments made on a review.

    If the review is a draft, then comments can be added, deleted, or
    changed on this list. However, if the review is already published,
    then no changes can be made.
    """
    allowed_methods = ('GET', 'POST', 'PUT', 'DELETE')
    policy_id = 'review_diff_comment'
    model_parent_key = 'review'
    uri_template_name = 'review_diff_comment'

    mimetype_list_resource_name = 'review-diff-comments'
    mimetype_item_resource_name = 'review-diff-comment'

    def get_queryset(self, request, review_id, *args, **kwargs):
        q = super(ReviewDiffCommentResource, self).get_queryset(
            request, *args, **kwargs)
        return q.filter(review=review_id)

    @webapi_check_local_site
    @webapi_login_required
    @webapi_response_errors(DOES_NOT_EXIST, INVALID_FORM_DATA,
                            NOT_LOGGED_IN, PERMISSION_DENIED)
    @webapi_request_fields(
        required=dict({
            'filediff_id': {
                'type': IntFieldType,
                'description': 'The ID of the file diff the comment is on.',
            },
            'first_line': {
                'type': IntFieldType,
                'description': 'The line number the comment starts at.',
            },
            'num_lines': {
                'type': IntFieldType,
                'description': 'The number of lines the comment spans.',
            },
        }, **BaseDiffCommentResource.REQUIRED_CREATE_FIELDS),
        optional=dict({
            'base_filediff_id': {
                'type': IntFieldType,
                'description': 'The ID of the base filediff for the '
                               ':term:`cumulative diff` the comment is on.\n'
                               '\n'
                               'This is only supported for review requests '
                               'created with commit history support.',
            },
            'interfilediff_id': {
                'type': IntFieldType,
                'description': 'The ID of the second file diff in the '
                               'interdiff the comment is on.',
            },
        }, **BaseDiffCommentResource.OPTIONAL_CREATE_FIELDS),
        allow_unknown=True,
    )
    def create(self, request, filediff_id, interfilediff_id=None,
               base_filediff_id=None, *args, **kwargs):
        """Creates a new diff comment.

        This will create a new diff comment on this review. The review
        must be a draft review.

        Extra data can be stored later lookup. See
        :ref:`webapi2.0-extra-data` for more information.
        """
        try:
            review_request = \
                resources.review_request.get_object(request, *args, **kwargs)
            review = resources.review.get_object(request, *args, **kwargs)
        except ObjectDoesNotExist:
            return DOES_NOT_EXIST

        if not resources.review.has_modify_permissions(request, review):
            return self.get_no_access_error(request)

        filediff = None
        interfilediff = None
        invalid_fields = {}

        try:
            query = (
                Q(diffset__history__review_request=review_request) |
                Q(diffset__review_request_draft__review_request=review_request))

            filediff = FileDiff.objects.get(
                Q(pk=filediff_id) & query)
        except ObjectDoesNotExist:
            invalid_fields['filediff_id'] = [
                'This is not a valid filediff ID.',
            ]

        if filediff is None or not dvcs_feature.is_enabled(request=request):
            base_filediff_id = None

        if base_filediff_id is not None:
            if not review_request.created_with_history:
                invalid_fields['base_filediff_id'] = [
                    'This field cannot be specified on review requests '
                    'created without history support.'
                ]
            elif interfilediff_id is not None:
                invalid_fields.update({
                    'base_filediff_id': [
                        'This field cannot be specified with '
                        'interfilediff_id.',
                    ],
                    'interfilediff_id': [
                        'This field cannot be specified with '
                        'base_filediff_id.',
                    ],
                })
            elif base_filediff_id == filediff_id:
                invalid_fields['base_filediff_id'] = [
                    'This cannot be the same as filediff_id.',
                ]

            elif base_filediff_id > filediff_id:
                invalid_fields['base_filediff_id'] = [
                    'This is not a valid base filediff ID.',
                ]
            else:
                base_filediff_exists = (
                    FileDiff.objects
                    .filter(diffset_id=filediff.diffset_id,
                            pk=base_filediff_id)
                    .exclude(commit_id=filediff.commit_id)
                    .exists()
                )

                if not base_filediff_exists:
                    invalid_fields['base_filediff_id'] = [
                        'This is not a valid base filediff ID.',
                    ]
                else:
                    ancestor_ids = (
                        ancestor.pk
                        for ancestor in filediff.get_ancestors(
                            minimal=False)
                    )

                    if base_filediff_id not in ancestor_ids:
                        invalid_fields['base_filediff_id'] = [
                            'This is not a valid base filediff ID.',
                        ]

        if filediff and interfilediff_id:
            if interfilediff_id == filediff.id:
                invalid_fields.setdefault('interfilediff_id', []).append(
                    'This cannot be the same as filediff_id.')
            else:
                try:
                    interfilediff = FileDiff.objects.get(
                        pk=interfilediff_id,
                        diffset__history=filediff.diffset.history)
                except ObjectDoesNotExist:
                    invalid_fields.setdefault('interfilediff_id', []).append(
                        'This is not a valid interfilediff ID.')

        if invalid_fields:
            return INVALID_FORM_DATA, {
                'fields': invalid_fields,
            }

        return self.create_comment(
            request=request,
            review=review,
            comments_m2m=review.comments,
            filediff=filediff,
            interfilediff=interfilediff,
            fields=('filediff', 'interfilediff', 'first_line', 'num_lines'),
            base_filediff_id=base_filediff_id,
            **kwargs)

    @webapi_check_local_site
    @webapi_login_required
    @webapi_response_errors(DOES_NOT_EXIST, NOT_LOGGED_IN, PERMISSION_DENIED)
    @webapi_request_fields(
        optional=dict({
            'first_line': {
                'type': IntFieldType,
                'description': 'The line number the comment starts at.',
            },
            'num_lines': {
                'type': IntFieldType,
                'description': 'The number of lines the comment spans.',
            },
        }, **BaseDiffCommentResource.OPTIONAL_UPDATE_FIELDS),
        allow_unknown=True,
    )
    def update(self, request, *args, **kwargs):
        """Updates a diff comment.

        This can update the text or line range of an existing comment.

        Extra data can be stored later lookup. See
        :ref:`webapi2.0-extra-data` for more information.
        """
        try:
            resources.review_request.get_object(request, *args, **kwargs)
            review = resources.review.get_object(request, *args, **kwargs)
            diff_comment = self.get_object(request, *args, **kwargs)
        except ObjectDoesNotExist:
            return DOES_NOT_EXIST

        return self.update_comment(request=request,
                                   review=review,
                                   comment=diff_comment,
                                   update_fields=('first_line', 'num_lines'),
                                   **kwargs)

    @augment_method_from(BaseDiffCommentResource)
    def delete(self, *args, **kwargs):
        """Deletes the comment.

        This will remove the comment from the review. This cannot be undone.

        Only comments on draft reviews can be deleted. Attempting to delete
        a published comment will return a Permission Denied error.

        Instead of a payload response, this will return :http:`204`.
        """
        pass

    @augment_method_from(BaseDiffCommentResource)
    def get_list(self, *args, **kwargs):
        """Returns the list of comments made on a review.

        This list can be filtered down by using the ``?line=`` and
        ``?interdiff-revision=``.

        To filter for comments that start on a particular line in the file,
        using ``?line=``.

        To filter for comments that span revisions of diffs, you can specify
        the second revision in the range using ``?interdiff-revision=``.
        """
        pass

    def create_comment(self, request, comments_m2m, base_filediff_id=None,
                       **kwargs):
        """Create a review comment.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            comments_m2m (django.db.models.ManyToManyField):
                The review's comments relation, where the new comment will be
                added.

            base_filediff_id (int, optional):
                The ID of the base filediff for the :term:`cumulative diff` the
                comment is on.

            **kwargs (dict):
                Additional keyword arguments to pass on to the base class
                method.

        Returns:
            tuple or djblets.webapi.errors.WebAPIError:
            Either a successful payload containing the comment, or an error
            payload.
        """
        rsp = super(ReviewDiffCommentResource, self).create_comment(
            comments_m2m=comments_m2m,
            save=False,
            **kwargs)

        if (isinstance(rsp, tuple) and
            isinstance(rsp[1], dict) and
            self.item_result_key in rsp[1]):
            comment = rsp[1][self.item_result_key]

            if (base_filediff_id is not None and
                dvcs_feature.is_enabled(request=request)):
                comment.base_filediff_id = base_filediff_id

            comment.save()
            comments_m2m.add(comment)

        return rsp

    def serialize_object(self, obj, request=None, *args, **kwargs):
        """Serialize a diff comment.

        Args:
            obj (reviewboard.reviews.models.diff_comment.Comment):
                The diff comment to serialize.

            request (django.http.HttpRequest, optional):
                The HTTP request from the client.

            *args (tuple):
                Additional positional arguments.

            **kwargs (dict):
                Additional keyword arguments.

        Returns:
            dict:
            The serialized diff comment.
        """
        result = super(ReviewDiffCommentResource, self).serialize_object(
            obj, request=request, *args, **kwargs)

        if not dvcs_feature.is_enabled(request=request):
            result.pop('base_filediff_id', None)

        return result


review_diff_comment_resource = ReviewDiffCommentResource()



"""Unit tests for reviewboard.certs.manager.CertificateManager.

Version Added:
    6.0
"""

from __future__ import annotations

import json
import os
import shutil
import ssl

import kgb
from django.conf import settings
from django.core.cache import cache
from djblets.cache.backend import make_cache_key
from djblets.siteconfig.models import SiteConfiguration

from reviewboard.certs.cert import (CertPurpose, CertificateBundle,
                                    CertificateFingerprints)
from reviewboard.certs.errors import CertificateNotFoundError
from reviewboard.certs.manager import CertificateManager, logger
from reviewboard.certs.storage import cert_storage_backend_registry
from reviewboard.certs.storage.file_storage import \
    FileCertificateStorageBackend
from reviewboard.certs.tests.testcases import (CaptureSSLContext,
                                               CertificateTestCase,
                                               TEST_CERT_BUNDLE_PEM,
                                               TEST_CLIENT_CERT_PEM,
                                               TEST_CLIENT_KEY_PEM,
                                               TEST_SHA1,
                                               TEST_SHA256,
                                               TEST_SHA256_2,
                                               TEST_TRUST_CERT_PEM)
from reviewboard.deprecation import RemovedInReviewBoard10_0Warning


class MyCertificateStorageBackend(FileCertificateStorageBackend):
    backend_id = 'test'


class CertificateManagerTests(kgb.SpyAgency, CertificateTestCase):
    """Unit tests for CertificateManager.

    Version Added:
        6.0
    """

    ######################
    # Instance variables #
    ######################

    #: The CertificateManager certs storage path for the test run.
    #:
    #: Type:
    #:     str
    certs_path: str

    def setUp(self) -> None:
        """Set up the test case.

        This will patch in a storage location for the certificate manager
        for this test run.
        """
        super().setUp()

        self.certs_path = os.path.join(settings.SITE_DATA_DIR, 'rb-certs')

    def tearDown(self) -> None:
        """Tear down the test case."""
        # Tests in this class mutate the cached SiteConfiguration singleton
        # without saving. Clear the key so it doesn't leak into other tests
        # sharing the same xdist worker.
        siteconfig = SiteConfiguration.objects.get_current()
        siteconfig.settings.pop('certs_storage_backend', None)

        if os.path.exists(self.certs_path):
            shutil.rmtree(self.certs_path)

        super().tearDown()

    def test_storage_backend_default(self) -> None:
        """Testing CertificateManager.storage_backend with default"""
        cert_manager = CertificateManager()
        storage_backend = cert_manager.storage_backend

        self.assertIsInstance(storage_backend, FileCertificateStorageBackend)
        self.assertEqual(storage_backend.storage_path,
                         os.path.join(self.certs_path, 'file'))

    def test_storage_backend_with_custom(self) -> None:
        """Testing CertificateManager.storage_backend with custom setting"""
        cert_storage_backend_registry.register(MyCertificateStorageBackend)

        self.addCleanup(
            lambda: cert_storage_backend_registry.unregister(
                MyCertificateStorageBackend))

        siteconfig = SiteConfiguration.objects.get_current()
        siteconfig.set('certs_storage_backend', 'test')

        cert_manager = CertificateManager()
        storage_backend = cert_manager.storage_backend

        self.assertIsInstance(storage_backend, MyCertificateStorageBackend)
        self.assertEqual(storage_backend.storage_path,
                         os.path.join(self.certs_path, 'test'))

    def test_storage_backend_after_setting_change(self) -> None:
        """Testing CertificateManager.storage_backend after setting is
        changed
        """
        cert_storage_backend_registry.register(MyCertificateStorageBackend)

        self.addCleanup(
            lambda: cert_storage_backend_registry.unregister(
                MyCertificateStorageBackend))

        cert_manager = CertificateManager()
        storage_backend = cert_manager.storage_backend

        self.assertIsInstance(storage_backend, FileCertificateStorageBackend)
        self.assertEqual(storage_backend.storage_path,
                         os.path.join(self.certs_path, 'file'))

        # Change the setting.
        siteconfig = SiteConfiguration.objects.get_current()
        siteconfig.set('certs_storage_backend', 'test')

        storage_backend = cert_manager.storage_backend

        self.assertIsInstance(storage_backend, MyCertificateStorageBackend)
        self.assertEqual(storage_backend.storage_path,
                         os.path.join(self.certs_path, 'test'))

    def test_storage_backend_with_missing_backend(self) -> None:
        """Testing CertificateManager.storage_backend with missing backend"""
        siteconfig = SiteConfiguration.objects.get_current()
        siteconfig.set('certs_storage_backend', 'test')

        cert_manager = CertificateManager()

        with self.assertLogs(logger) as cm:
            storage_backend = cert_manager.storage_backend

        self.assertEqual(
            cm.records[0].getMessage(),
            'Unable to load SSL/TLS certificate storage backend "test". '
            'Falling back to file-based storage.')

        self.assertIsInstance(storage_backend, FileCertificateStorageBackend)
        self.assertEqual(storage_backend.storage_path,
                         os.path.join(self.certs_path, 'file'))

    def test_add_ca_bundle(self) -> None:
        """Testing CertificateManager.add_ca_bundle"""
        cert_manager = CertificateManager()
        storage_backend = cert_manager.storage_backend
        path = os.path.join(storage_backend.storage_path, 'cabundles',
                            'my-certs.pem')

        stored_ca_bundle = cert_manager.add_ca_bundle(
            CertificateBundle(bundle_data=TEST_CERT_BUNDLE_PEM,
                              name='my-certs'))

        self.assertIsNotNone(stored_ca_bundle)
        self.assertAttrsEqual(
            stored_ca_bundle,
            {
                'local_site': None,
                'storage': storage_backend,
            })
        self.assertTrue(os.path.exists(path))
        self.assertEqual(stored_ca_bundle.get_bundle_file_path(), path)

        with open(path, 'rb') as fp:
            self.assertEqual(fp.read(), TEST_CERT_BUNDLE_PEM)

        self.assertAttrsEqual(
            stored_ca_bundle.bundle,
            {
                'bundle_data': TEST_CERT_BUNDLE_PEM,
                'name': 'my-certs',
            })

    def test_add_ca_bundle_with_local_site(self) -> None:
        """Testing CertificateManager.add_ca_bundle with LocalSite"""
        local_site = self.create_local_site('test-site')

        cert_manager = CertificateManager()
        storage_backend = cert_manager.storage_backend
        path = os.path.join(storage_backend.storage_path, 'sites',
                            'test-site', 'cabundles', 'my-certs.pem')

        stored_ca_bundle = cert_manager.add_ca_bundle(
            CertificateBundle(bundle_data=TEST_CERT_BUNDLE_PEM,
                              name='my-certs'),
            local_site=local_site)

        self.assertIsNotNone(stored_ca_bundle)
        self.assertAttrsEqual(
            stored_ca_bundle,
            {
                'local_site': local_site,
                'storage': storage_backend,
            })
        self.assertTrue(os.path.exists(path))
        self.assertEqual(stored_ca_bundle.get_bundle_file_path(), path)

        with open(path, 'rb') as fp:
            self.assertEqual(fp.read(), TEST_CERT_BUNDLE_PEM)

        self.assertAttrsEqual(
            stored_ca_bundle.bundle,
            {
                'bundle_data': TEST_CERT_BUNDLE_PEM,
                'name': 'my-certs',
            })

    def test_delete_ca_bundle(self) -> None:
        """Testing CertificateManager.delete_ca_bundle"""
        cert_manager = CertificateManager()
        storage_backend = cert_manager.storage_backend
        path = os.path.join(storage_backend.storage_path, 'cabundles',
                            'my-certs.pem')

        cert_manager.add_ca_bundle(
            CertificateBundle(bundle_data=TEST_CERT_BUNDLE_PEM,
                              name='my-certs'))
        self.assertTrue(os.path.exists(path))

        storage_backend.delete_ca_bundle(name='my-certs')
        self.assertFalse(os.path.exists(path))

    def test_delete_ca_bundle_with_local_site(self) -> None:
        """Testing CertificateManager.delete_ca_bundle with LocalSite"""
        local_site = self.create_local_site('test-site')

        cert_manager = CertificateManager()
        storage_backend = cert_manager.storage_backend
        path = os.path.join(storage_backend.storage_path, 'sites',
                            'test-site', 'cabundles', 'my-certs.pem')

        cert_manager.add_ca_bundle(
            CertificateBundle(bundle_data=TEST_CERT_BUNDLE_PEM,
                              name='my-certs'),
            local_site=local_site)
        self.assertTrue(os.path.exists(path))

        storage_backend.delete_ca_bundle(name='my-certs',
                                         local_site=local_site)
        self.assertFalse(os.path.exists(path))

    def test_delete_ca_bundle_with_local_site_mismatch(self) -> None:
        """Testing CertificateManager.delete_ca_bundle with LocalSite
        mismatch
        """
        local_site1 = self.create_local_site('test-site-1')
        local_site2 = self.create_local_site('test-site-2')

        cert_manager = CertificateManager()
        storage_backend = cert_manager.storage_backend
        path = os.path.join(storage_backend.storage_path, 'sites',
                            'test-site-1', 'cabundles', 'my-certs.pem')

        cert_manager.add_ca_bundle(
            CertificateBundle(bundle_data=TEST_CERT_BUNDLE_PEM,
                              name='my-certs'),
            local_site=local_site1)
        self.assertTrue(os.path.exists(path))

        message = 'The SSL/TLS certificate was not found.'

        with self.assertRaisesMessage(CertificateNotFoundError, message):
            storage_backend.delete_ca_bundle(name='my-certs')

        with self.assertRaisesMessage(CertificateNotFoundError, message):
            storage_backend.delete_ca_bundle(name='my-certs',
                                             local_site=local_site2)

        self.assertTrue(os.path.exists(path))

    def test_get_ca_bundle(self) -> None:
        """Testing CertificateManager.get_ca_bundle"""
        local_site = self.create_local_site('test-site-1')

        cert_manager = CertificateManager()

        cert_manager.add_ca_bundle(
            CertificateBundle(bundle_data=TEST_CERT_BUNDLE_PEM,
                              name='my-certs'))
        ca_bundle = cert_manager.get_ca_bundle(name='my-certs')
        assert ca_bundle is not None

        self.assertAttrsEqual(
            ca_bundle,
            {
                'bundle_data': TEST_CERT_BUNDLE_PEM,
                'name': 'my-certs',
            })

        # Make sure it's only accessible on the global site.
        self.assertIsNone(cert_manager.get_ca_bundle(name='my-certs',
                                                     local_site=local_site))

    def test_get_ca_bundle_with_local_site(self) -> None:
        """Testing CertificateManager.get_ca_bundle with LocalSite"""
        local_site1 = self.create_local_site('test-site-1')
        local_site2 = self.create_local_site('test-site-2')

        cert_manager = CertificateManager()
        cert_manager.add_ca_bundle(
            CertificateBundle(bundle_data=TEST_CERT_BUNDLE_PEM,
                              name='my-certs'),
            local_site=local_site1)
        ca_bundle = cert_manager.get_ca_bundle(name='my-certs',
                                               local_site=local_site1)
        assert ca_bundle is not None

        self.assertAttrsEqual(
            ca_bundle,
            {
                'bundle_data': TEST_CERT_BUNDLE_PEM,
                'name': 'my-certs',
            })

        # Make sure it's only accessible on local_site1.
        self.assertIsNone(cert_manager.get_ca_bundle(name='my-certs'))
        self.assertIsNone(cert_manager.get_ca_bundle(name='my-certs',
                                                     local_site=local_site2))

    def test_get_ca_bundle_with_not_found(self) -> None:
        """Testing CertificateManager.get_ca_bundle with not found"""
        cert_manager = CertificateManager()

        self.assertIsNone(cert_manager.get_ca_bundle(name='missing-certs'))

    def test_get_ca_bundles_dir(self) -> None:
        """Testing CertificateManager.get_ca_bundles_dir"""
        cert_manager = CertificateManager()

        self.assertEqual(cert_manager.get_ca_bundles_dir(),
                         os.path.join(self.certs_path, 'file', 'cabundles'))

    def test_get_ca_bundles_dir_with_local_site(self) -> None:
        """Testing CertificateManager.get_ca_bundles_dir"""
        local_site = self.create_local_site('test-site-1')

        cert_manager = CertificateManager()

        self.assertEqual(
            cert_manager.get_ca_bundles_dir(local_site=local_site),
            os.path.join(self.certs_path, 'file', 'sites', 'test-site-1',
                         'cabundles'))

    def test_add_certificate_with_trust(self) -> None:
        """Testing CertificateManager.add_certificate with purpose=trust"""
        cert_manager = CertificateManager()
        storage_backend = cert_manager.storage_backend

        cert_path = os.path.join(storage_backend.storage_path, 'certs',
                                 'trust', 'example.com__443.crt')
        key_path = os.path.join(storage_backend.storage_path, 'certs',
                                'trust', 'example.com__443.key')
        fingerprints_path = os.path.join(storage_backend.storage_path,
                                         'fingerprints',
                                         'example.com__443.json')

        stored_cert = cert_manager.add_certificate(
            self.create_certificate(cert_data=TEST_TRUST_CERT_PEM))

        self.assertIsNotNone(stored_cert)
        self.assertAttrsEqual(
            stored_cert,
            {
                'local_site': None,
                'purpose': CertPurpose.TRUST,
                'storage': storage_backend,
            })
        self.assertTrue(os.path.exists(cert_path))
        self.assertFalse(os.path.exists(key_path))
        self.assertTrue(os.path.exists(fingerprints_path))
        self.assertEqual(stored_cert.get_cert_file_path(), cert_path)
        self.assertIsNone(stored_cert.get_key_file_path())

        with open(cert_path, 'rb') as fp:
            self.assertEqual(fp.read(), TEST_TRUST_CERT_PEM)

        with open(fingerprints_path, 'rb') as fp:
            self.assertEqual(json.load(fp), {
                'sha1': TEST_SHA1,
                'sha256': TEST_SHA256,
            })

        self.assertAttrsEqual(
            stored_cert.certificate,
            {
                'cert_data': TEST_TRUST_CERT_PEM,
                'hostname': 'example.com',
                'key_data': None,
                'port': 443,
                'purpose': CertPurpose.TRUST,
            })

    def test_add_certificate_with_client(self) -> None:
        """Testing CertificateManager.add_certificate with purpose=client"""
        cert_manager = CertificateManager()
        storage_backend = cert_manager.storage_backend

        cert_path = os.path.join(storage_backend.storage_path, 'certs',
                                 'client', 'example.com__443.crt')
        key_path = os.path.join(storage_backend.storage_path, 'certs',
                                'client', 'example.com__443.key')

        stored_cert = cert_manager.add_certificate(
            self.create_certificate(purpose=CertPurpose.CLIENT,
                                    cert_data=TEST_CLIENT_CERT_PEM,
                                    key_data=TEST_CLIENT_KEY_PEM))

        self.assertIsNotNone(stored_cert)
        self.assertAttrsEqual(
            stored_cert,
            {
                'local_site': None,
                'purpose': CertPurpose.CLIENT,
                'storage': storage_backend,
            })
        self.assertTrue(os.path.exists(cert_path))
        self.assertTrue(os.path.exists(key_path))
        self.assertEqual(stored_cert.get_cert_file_path(), cert_path)
        self.assertEqual(stored_cert.get_key_file_path(), key_path)

        # Fingerprints should not exist for these certificates.
        self.assertFalse(os.path.exists(os.path.join(
            storage_backend.storage_path, 'fingerprints',
            'example.com__443.json')))

        with open(cert_path, 'rb') as fp:
            self.assertEqual(fp.read(), TEST_CLIENT_CERT_PEM)

        with open(key_path, 'rb') as fp:
            self.assertEqual(fp.read(), TEST_CLIENT_KEY_PEM)

        self.assertAttrsEqual(
            stored_cert.certificate,
            {
                'cert_data': TEST_CLIENT_CERT_PEM,
                'hostname': 'example.com',
                'key_data': TEST_CLIENT_KEY_PEM,
                'port': 443,
                'purpose': CertPurpose.CLIENT,
            })

    def test_add_certificate_with_trust_and_local_site(self) -> None:
        """Testing CertificateManager.add_certificate with purpose=trust and
        LocalSite
        """
        local_site = self.create_local_site('test-site-1')

        cert_manager = CertificateManager()
        storage_backend = cert_manager.storage_backend

        site_storage_path = os.path.join(storage_backend.storage_path, 'sites',
                                         'test-site-1')
        cert_path = os.path.join(site_storage_path, 'certs', 'trust',
                                 'example.com__443.crt')
        key_path = os.path.join(site_storage_path, 'certs', 'trust',
                                'example.com__443.key')
        fingerprints_path = os.path.join(site_storage_path, 'fingerprints',
                                         'example.com__443.json')

        stored_cert = cert_manager.add_certificate(
            self.create_certificate(cert_data=TEST_TRUST_CERT_PEM),
            local_site=local_site)

        self.assertIsNotNone(stored_cert)
        self.assertAttrsEqual(
            stored_cert,
            {
                'local_site': local_site,
                'purpose': CertPurpose.TRUST,
                'storage': storage_backend,
            })
        self.assertTrue(os.path.exists(cert_path))
        self.assertFalse(os.path.exists(key_path))
        self.assertTrue(os.path.exists(fingerprints_path))
        self.assertEqual(stored_cert.get_cert_file_path(), cert_path)
        self.assertIsNone(stored_cert.get_key_file_path())

        with open(cert_path, 'rb') as fp:
            self.assertEqual(fp.read(), TEST_TRUST_CERT_PEM)

        with open(fingerprints_path, 'rb') as fp:
            self.assertEqual(json.load(fp), {
                'sha1': TEST_SHA1,
                'sha256': TEST_SHA256,
            })

        self.assertAttrsEqual(
            stored_cert.certificate,
            {
                'cert_data': TEST_TRUST_CERT_PEM,
                'hostname': 'example.com',
                'key_data': None,
                'port': 443,
                'purpose': CertPurpose.TRUST,
            })

    def test_add_certificate_with_client_and_local_site(self) -> None:
        """Testing CertificateManager.add_certificate with purpose=client and
        LocalSite
        """
        local_site = self.create_local_site('test-site-1')

        cert_manager = CertificateManager()
        storage_backend = cert_manager.storage_backend

        site_storage_path = os.path.join(storage_backend.storage_path, 'sites',
                                         'test-site-1')
        cert_path = os.path.join(site_storage_path, 'certs', 'client',
                                 'example.com__443.crt')
        key_path = os.path.join(site_storage_path, 'certs', 'client',
                                'example.com__443.key')

        stored_cert = cert_manager.add_certificate(
            self.create_certificate(purpose=CertPurpose.CLIENT,
                                    cert_data=TEST_CLIENT_CERT_PEM,
                                    key_data=TEST_CLIENT_KEY_PEM),
            local_site=local_site)

        self.assertIsNotNone(stored_cert)
        self.assertAttrsEqual(
            stored_cert,
            {
                'local_site': local_site,
                'purpose': CertPurpose.CLIENT,
                'storage': storage_backend,
            })
        self.assertTrue(os.path.exists(cert_path))
        self.assertTrue(os.path.exists(key_path))
        self.assertEqual(stored_cert.get_cert_file_path(), cert_path)
        self.assertEqual(stored_cert.get_key_file_path(), key_path)

        # Fingerprints should not exist for these certificates.
        self.assertFalse(os.path.exists(os.path.join(
            storage_backend.storage_path, 'fingerprints',
            'example.com__443.json')))

        with open(cert_path, 'rb') as fp:
            self.assertEqual(fp.read(), TEST_CLIENT_CERT_PEM)

        with open(key_path, 'rb') as fp:
            self.assertEqual(fp.read(), TEST_CLIENT_KEY_PEM)

        self.assertAttrsEqual(
            stored_cert.certificate,
            {
                'cert_data': TEST_CLIENT_CERT_PEM,
                'hostname': 'example.com',
                'key_data': TEST_CLIENT_KEY_PEM,
                'port': 443,
                'purpose': CertPurpose.CLIENT,
            })

    def test_delete_certificate_with_trust(self) -> None:
        """Testing CertificateManager.delete_certificate with purpose=trust"""
        cert_manager = CertificateManager()
        storage_backend = cert_manager.storage_backend

        cert_path = os.path.join(storage_backend.storage_path, 'certs',
                                 'trust', 'example.com__443.crt')
        key_path = os.path.join(storage_backend.storage_path, 'certs',
                                'trust', 'example.com__443.key')
        fingerprints_path = os.path.join(storage_backend.storage_path,
                                         'fingerprints',
                                         'example.com__443.json')

        cert_manager.add_certificate(
            self.create_certificate(cert_data=TEST_TRUST_CERT_PEM))

        self.assertTrue(os.path.exists(cert_path))
        self.assertFalse(os.path.exists(key_path))
        self.assertTrue(os.path.exists(fingerprints_path))

        cert_manager.delete_certificate(hostname='example.com',
                                        port=443)

        self.assertFalse(os.path.exists(cert_path))
        self.assertFalse(os.path.exists(key_path))
        self.assertFalse(os.path.exists(fingerprints_path))

    def test_delete_certificate_with_client(self) -> None:
        """Testing CertificateManager.delete_certificate with purpose=client"""
        cert_manager = CertificateManager()
        storage_backend = cert_manager.storage_backend

        cert_path = os.path.join(storage_backend.storage_path, 'certs',
                                 'client', 'example.com__443.crt')
        key_path = os.path.join(storage_backend.storage_path, 'certs',
                                'client', 'example.com__443.key')
        fingerprints_path = os.path.join(storage_backend.storage_path,
                                         'fingerprints',
                                         'example.com__443.json')

        cert_manager.add_certificate(
            self.create_certificate(purpose=CertPurpose.CLIENT,
                                    cert_data=TEST_CLIENT_CERT_PEM,
                                    key_data=TEST_CLIENT_KEY_PEM))

        self.assertTrue(os.path.exists(cert_path))
        self.assertTrue(os.path.exists(key_path))
        self.assertFalse(os.path.exists(fingerprints_path))

        cert_manager.delete_certificate(hostname='example.com',
                                        port=443,
                                        purpose=CertPurpose.CLIENT)

        self.assertFalse(os.path.exists(cert_path))
        self.assertFalse(os.path.exists(key_path))
        self.assertFalse(os.path.exists(fingerprints_path))

    def test_delete_certificate_with_trust_and_local_site(self) -> None:
        """Testing CertificateManager.delete_certificate with purpose=trust
        and LocalSite
        """
        local_site = self.create_local_site('test-site-1')

        cert_manager = CertificateManager()
        storage_backend = cert_manager.storage_backend

        site_storage_path = os.path.join(storage_backend.storage_path, 'sites',
                                         'test-site-1')
        cert_path = os.path.join(site_storage_path, 'certs', 'trust',
                                 'example.com__443.crt')
        key_path = os.path.join(site_storage_path, 'certs', 'trust',
                                'example.com__443.key')
        fingerprints_path = os.path.join(site_storage_path, 'fingerprints',
                                         'example.com__443.json')

        cert_manager.add_certificate(
            self.create_certificate(cert_data=TEST_TRUST_CERT_PEM),
            local_site=local_site)

        self.assertTrue(os.path.exists(cert_path))
        self.assertFalse(os.path.exists(key_path))
        self.assertTrue(os.path.exists(fingerprints_path))

        cert_manager.delete_certificate(hostname='example.com',
                                        port=443,
                                        local_site=local_site)

        self.assertFalse(os.path.exists(cert_path))
        self.assertFalse(os.path.exists(key_path))
        self.assertFalse(os.path.exists(fingerprints_path))

    def test_delete_certificate_with_client_and_local_site(self) -> None:
        """Testing CertificateManager.delete_certificate with purpose=client
        and LocalSite
        """
        local_site = self.create_local_site('test-site-1')

        cert_manager = CertificateManager()
        storage_backend = cert_manager.storage_backend

        site_storage_path = os.path.join(storage_backend.storage_path, 'sites',
                                         'test-site-1')
        cert_path = os.path.join(site_storage_path, 'certs', 'client',
                                 'example.com__443.crt')
        key_path = os.path.join(site_storage_path, 'certs', 'client',
                                'example.com__443.key')
        fingerprints_path = os.path.join(site_storage_path, 'fingerprints',
                                         'example.com__443.json')

        cert_manager.add_certificate(
            self.create_certificate(purpose=CertPurpose.CLIENT,
                                    cert_data=TEST_CLIENT_CERT_PEM,
                                    key_data=TEST_CLIENT_KEY_PEM),
            local_site=local_site)

        self.assertTrue(os.path.exists(cert_path))
        self.assertTrue(os.path.exists(key_path))
        self.assertFalse(os.path.exists(fingerprints_path))

        cert_manager.delete_certificate(hostname='example.com',
                                        port=443,
                                        purpose=CertPurpose.CLIENT,
                                        local_site=local_site)

        self.assertFalse(os.path.exists(cert_path))
        self.assertFalse(os.path.exists(key_path))
        self.assertFalse(os.path.exists(fingerprints_path))

    def test_delete_certificate_with_trust_and_local_site_mismatch(
        self,
    ) -> None:
        """Testing CertificateManager.delete_certificate with purpose=trust
        and LocalSite mismatch
        """
        local_site1 = self.create_local_site('test-site-1')
        local_site2 = self.create_local_site('test-site-2')

        cert_manager = CertificateManager()
        storage_backend = cert_manager.storage_backend

        site_storage_path = os.path.join(storage_backend.storage_path, 'sites',
                                         'test-site-1')
        cert_path = os.path.join(site_storage_path, 'certs', 'trust',
                                 'example.com__443.crt')
        key_path = os.path.join(site_storage_path, 'certs', 'trust',
                                'example.com__443.key')
        fingerprints_path = os.path.join(site_storage_path, 'fingerprints',
                                         'example.com__443.json')

        cert_manager.add_certificate(
            self.create_certificate(cert_data=TEST_TRUST_CERT_PEM),
            local_site=local_site1)

        message = 'The SSL/TLS certificate was not found.'

        with self.assertRaisesMessage(CertificateNotFoundError, message):
            cert_manager.delete_certificate(hostname='example.com',
                                            port=443)

        with self.assertRaisesMessage(CertificateNotFoundError, message):
            cert_manager.delete_certificate(hostname='example.com',
                                            port=443,
                                            local_site=local_site2)

        self.assertTrue(os.path.exists(cert_path))
        self.assertFalse(os.path.exists(key_path))
        self.assertTrue(os.path.exists(fingerprints_path))

    def test_delete_certificate_with_client_and_local_site_mismatch(
        self,
    ) -> None:
        """Testing CertificateManager.delete_certificate with purpose=client
        and LocalSite mismatch
        """
        local_site1 = self.create_local_site('test-site-1')
        local_site2 = self.create_local_site('test-site-2')

        cert_manager = CertificateManager()
        storage_backend = cert_manager.storage_backend

        site_storage_path = os.path.join(storage_backend.storage_path, 'sites',
                                         'test-site-1')
        cert_path = os.path.join(site_storage_path, 'certs', 'client',
                                 'example.com__443.crt')
        key_path = os.path.join(site_storage_path, 'certs', 'client',
                                'example.com__443.key')
        fingerprints_path = os.path.join(site_storage_path, 'fingerprints',
                                         'example.com__443.json')

        cert_manager.add_certificate(
            self.create_certificate(purpose=CertPurpose.CLIENT,
                                    cert_data=TEST_CLIENT_CERT_PEM,
                                    key_data=TEST_CLIENT_KEY_PEM),
            local_site=local_site1)

        message = 'The SSL/TLS certificate was not found.'

        with self.assertRaisesMessage(CertificateNotFoundError, message):
            cert_manager.delete_certificate(hostname='example.com',
                                            port=443)

        with self.assertRaisesMessage(CertificateNotFoundError, message):
            cert_manager.delete_certificate(hostname='example.com',
                                            port=443,
                                            local_site=local_site2)

        self.assertTrue(os.path.exists(cert_path))
        self.assertTrue(os.path.exists(key_path))
        self.assertFalse(os.path.exists(fingerprints_path))

    def test_get_certificate_with_trust(self) -> None:
        """Testing CertificateManager.get_certificate with purpose=trust"""
        local_site = self.create_local_site('test-site-1')

        cert_manager = CertificateManager()

        cert_manager.add_certificate(
            self.create_certificate(cert_data=TEST_TRUST_CERT_PEM))
        cert_manager.add_certificate(
            self.create_certificate(purpose=CertPurpose.CLIENT,
                                    cert_data=TEST_CLIENT_CERT_PEM,
                                    key_data=TEST_CLIENT_KEY_PEM))

        certificate = cert_manager.get_certificate(
            hostname='example.com',
            port=443,
            purpose=CertPurpose.TRUST,
        )

        assert certificate is not None

        self.assertAttrsEqual(
            certificate,
            {
                'cert_data': TEST_TRUST_CERT_PEM,
                'hostname': 'example.com',
                'key_data': None,
                'purpose': CertPurpose.TRUST,
                'port': 443,
            })

        # Make sure it's only accessible on the global site.
        self.assertIsNone(cert_manager.get_certificate(hostname='example.com',
                                                       port=443,
                                                       local_site=local_site))

    def test_get_certificate_with_client(self) -> None:
        """Testing CertificateManager.get_certificate with purpose=client"""
        local_site = self.create_local_site('test-site-1')

        cert_manager = CertificateManager()

        cert_manager.add_certificate(
            self.create_certificate(cert_data=TEST_TRUST_CERT_PEM))
        cert_manager.add_certificate(
            self.create_certificate(purpose=CertPurpose.CLIENT,
                                    cert_data=TEST_CLIENT_CERT_PEM,
                                    key_data=TEST_CLIENT_KEY_PEM))

        certificate = cert_manager.get_certificate(
            hostname='example.com',
            port=443,
            purpose=CertPurpose.CLIENT,
        )

        assert certificate is not None

        self.assertAttrsEqual(
            certificate,
            {
                'cert_data': TEST_CLIENT_CERT_PEM,
                'hostname': 'example.com',
                'key_data': TEST_CLIENT_KEY_PEM,
                'port': 443,
                'purpose': CertPurpose.CLIENT,
            })

        # Make sure it's only accessible on the global site.
        self.assertIsNone(cert_manager.get_certificate(
            hostname='example.com',
            port=443,
            local_site=local_site,
            purpose=CertPurpose.CLIENT,
        ))

    def test_get_certificate_with_trust_and_local_site(self) -> None:
        """Testing CertificateManager.get_certificate with purpose=trust
        and LocalSite
        """
        local_site1 = self.create_local_site('test-site-1')
        local_site2 = self.create_local_site('test-site-2')

        cert_manager = CertificateManager()

        cert_manager.add_certificate(
            self.create_certificate(cert_data=TEST_TRUST_CERT_PEM),
            local_site=local_site1,
        )
        cert_manager.add_certificate(
            self.create_certificate(
                purpose=CertPurpose.CLIENT,
                cert_data=TEST_CLIENT_CERT_PEM,
                key_data=TEST_CLIENT_KEY_PEM,
            ),
            local_site=local_site1,
        )

        certificate = cert_manager.get_certificate(
            hostname='example.com',
            port=443,
            local_site=local_site1,
        )

        assert certificate is not None

        self.assertAttrsEqual(
            certificate,
            {
                'cert_data': TEST_TRUST_CERT_PEM,
                'hostname': 'example.com',
                'key_data': None,
                'port': 443,
                'purpose': CertPurpose.TRUST,
            })

        # Make sure it's only accessible on local_site1.
        self.assertIsNone(cert_manager.get_certificate(hostname='example.com',
                                                       port=443))
        self.assertIsNone(cert_manager.get_certificate(hostname='example.com',
                                                       port=443,
                                                       local_site=local_site2))

    def test_get_certificate_with_client_and_local_site(self) -> None:
        """Testing CertificateManager.get_certificate with purpose=client
        and LocalSite
        """
        local_site1 = self.create_local_site('test-site-1')
        local_site2 = self.create_local_site('test-site-2')

        cert_manager = CertificateManager()

        cert_manager.add_certificate(
            self.create_certificate(cert_data=TEST_TRUST_CERT_PEM),
            local_site=local_site1,
        )
        cert_manager.add_certificate(
            self.create_certificate(
                purpose=CertPurpose.CLIENT,
                cert_data=TEST_CLIENT_CERT_PEM,
                key_data=TEST_CLIENT_KEY_PEM,
            ),
            local_site=local_site1,
        )

        certificate = cert_manager.get_certificate(
            hostname='example.com',
            port=443,
            purpose=CertPurpose.CLIENT,
            local_site=local_site1,
        )

        assert certificate is not None

        self.assertAttrsEqual(
            certificate,
            {
                'cert_data': TEST_CLIENT_CERT_PEM,
                'hostname': 'example.com',
                'key_data': TEST_CLIENT_KEY_PEM,
                'port': 443,
                'purpose': CertPurpose.CLIENT,
            })

        # Make sure it's only accessible on local_site1.
        self.assertIsNone(cert_manager.get_certificate(
            hostname='example.com',
            port=443,
            purpose=CertPurpose.CLIENT,
        ))
        self.assertIsNone(cert_manager.get_certificate(
            hostname='example.com',
            port=443,
            local_site=local_site2,
            purpose=CertPurpose.CLIENT,
        ))

    def test_get_certificate_with_not_found(self) -> None:
        """Testing CertificateManager.get_certificate with not found"""
        cert_manager = CertificateManager()

        self.assertIsNone(cert_manager.get_certificate(hostname='example.com',
                                                       port=443,))

    def test_get_certificate_file_paths_with_trust(self) -> None:
        """Testing CertificateManager.get_certificate_file_paths with
        purpose=trust
        """
        local_site = self.create_local_site('test-site-1')

        cert_manager = CertificateManager()
        storage_backend = cert_manager.storage_backend

        cert_manager.add_certificate(
            self.create_certificate(cert_data=TEST_TRUST_CERT_PEM))

        self.assertEqual(
            cert_manager.get_certificate_file_paths(hostname='example.com',
                                                    port=443),
            {
                'cert_file': os.path.join(storage_backend.storage_path,
                                          'certs', 'trust',
                                          'example.com__443.crt'),
            })

        # Make sure it's only accessible on the global site.
        self.assertIsNone(cert_manager.get_certificate_file_paths(
            hostname='example.com',
            port=443,
            local_site=local_site,
        ))

    def test_get_certificate_file_paths_with_client(self) -> None:
        """Testing CertificateManager.get_certificate_file_paths with
        purpose=client
        """
        local_site = self.create_local_site('test-site-1')

        cert_manager = CertificateManager()
        storage_backend = cert_manager.storage_backend

        cert_manager.add_certificate(self.create_certificate(
            cert_data=TEST_TRUST_CERT_PEM,
        ))
        cert_manager.add_certificate(self.create_certificate(
            purpose=CertPurpose.CLIENT,
            cert_data=TEST_CLIENT_CERT_PEM,
            key_data=TEST_CLIENT_KEY_PEM,
        ))

        self.assertEqual(
            cert_manager.get_certificate_file_paths(
                hostname='example.com',
                port=443,
                purpose=CertPurpose.CLIENT,
            ),
            {
                'cert_file': os.path.join(storage_backend.storage_path,
                                          'certs', 'client',
                                          'example.com__443.crt'),
                'key_file': os.path.join(storage_backend.storage_path,
                                         'certs', 'client',
                                         'example.com__443.key'),
            })

        # Make sure it's only accessible on the global site.
        self.assertIsNone(cert_manager.get_certificate_file_paths(
            hostname='example.com',
            port=443,
            purpose=CertPurpose.CLIENT,
            local_site=local_site,
        ))

    def test_get_certificate_file_paths_with_trust_and_local_site(
        self,
    ) -> None:
        """Testing CertificateManager.get_certificate_file_paths with
        purpose=trust and LocalSite
        """
        local_site1 = self.create_local_site('test-site-1')
        local_site2 = self.create_local_site('test-site-2')

        cert_manager = CertificateManager()
        storage_backend = cert_manager.storage_backend
        site_storage_path = os.path.join(storage_backend.storage_path, 'sites',
                                         'test-site-1')

        cert_manager.add_certificate(
            self.create_certificate(cert_data=TEST_TRUST_CERT_PEM),
            local_site=local_site1,
        )
        cert_manager.add_certificate(
            self.create_certificate(
                purpose=CertPurpose.CLIENT,
                cert_data=TEST_CLIENT_CERT_PEM,
                key_data=TEST_CLIENT_KEY_PEM,
            ),
            local_site=local_site1,
        )

        self.assertEqual(
            cert_manager.get_certificate_file_paths(
                hostname='example.com',
                port=443,
                local_site=local_site1,
            ),
            {
                'cert_file': os.path.join(site_storage_path, 'certs',
                                          'trust', 'example.com__443.crt'),
            })

        # Make sure it's only accessible on local_site1.
        self.assertIsNone(cert_manager.get_certificate_file_paths(
            hostname='example.com',
            port=443,
        ))
        self.assertIsNone(cert_manager.get_certificate_file_paths(
            hostname='example.com',
            port=443,
            local_site=local_site2,
        ))

    def test_get_certificate_file_paths_with_client_and_local_site(
        self,
    ) -> None:
        """Testing CertificateManager.get_certificate_file_paths with
        purpose=client and LocalSite
        """
        local_site1 = self.create_local_site('test-site-1')
        local_site2 = self.create_local_site('test-site-2')

        cert_manager = CertificateManager()
        storage_backend = cert_manager.storage_backend
        site_storage_path = os.path.join(storage_backend.storage_path, 'sites',
                                         'test-site-1')

        cert_manager.add_certificate(
            self.create_certificate(cert_data=TEST_TRUST_CERT_PEM),
            local_site=local_site1,
        )
        cert_manager.add_certificate(
            self.create_certificate(
                purpose=CertPurpose.CLIENT,
                cert_data=TEST_CLIENT_CERT_PEM,
                key_data=TEST_CLIENT_KEY_PEM,
            ),
            local_site=local_site1,
        )

        self.assertEqual(
            cert_manager.get_certificate_file_paths(
                hostname='example.com',
                port=443,
                purpose=CertPurpose.CLIENT,
                local_site=local_site1,
            ),
            {
                'cert_file': os.path.join(site_storage_path, 'certs',
                                          'client', 'example.com__443.crt'),
                'key_file': os.path.join(site_storage_path, 'certs',
                                         'client', 'example.com__443.key'),
            })

        # Make sure it's only accessible on local_site1.
        self.assertIsNone(cert_manager.get_certificate_file_paths(
            hostname='example.com',
            port=443,
            purpose=CertPurpose.CLIENT,
        ))
        self.assertIsNone(cert_manager.get_certificate_file_paths(
            hostname='example.com',
            port=443,
            local_site=local_site2,
            purpose=CertPurpose.CLIENT,
        ))

    def test_mark_certificate_verified(self) -> None:
        """Testing CertificateManager.mark_certificate_verified"""
        cert_manager = CertificateManager()
        storage_backend = cert_manager.storage_backend
        fingerprints_path = os.path.join(storage_backend.storage_path,
                                         'fingerprints',
                                         'example.com__443.json')

        cert_manager.mark_certificate_verified(
            hostname='example.com',
            port=443,
            fingerprints=self.create_certificate_fingerprints())

        self.assertTrue(os.path.exists(fingerprints_path))

        with open(fingerprints_path, 'rb') as fp:
            self.assertEqual(json.load(fp), {
                'sha1': TEST_SHA1,
                'sha256': TEST_SHA256,
            })

        # This should also be in cache.
        self.assertEqual(
            cache.get(make_cache_key(
                'rb-ssl-fingerprints:file:example.com:443')),
            {
                'sha1': TEST_SHA1,
                'sha256': TEST_SHA256,
            })

    def test_mark_certificate_verified_with_local_site(self) -> None:
        """Testing CertificateManager.mark_certificate_verified with LocalSite
        """
        local_site = self.create_local_site('test-site-1')

        cert_manager = CertificateManager()
        storage_backend = cert_manager.storage_backend
        fingerprints_path = os.path.join(storage_backend.storage_path,
                                         'sites', 'test-site-1',
                                         'fingerprints',
                                         'example.com__443.json')

        cert_manager.mark_certificate_verified(
            hostname='example.com',
            port=443,
            fingerprints=self.create_certificate_fingerprints(),
            local_site=local_site)

        self.assertTrue(os.path.exists(fingerprints_path))

        with open(fingerprints_path, 'rb') as fp:
            self.assertEqual(json.load(fp), {
                'sha1': TEST_SHA1,
                'sha256': TEST_SHA256,
            })

        # This should also be in cache.
        self.assertEqual(
            cache.get(make_cache_key(
                'rb-ssl-fingerprints:file:1:example.com:443')),
            {
                'sha1': TEST_SHA1,
                'sha256': TEST_SHA256,
            })

    def test_mark_certificate_verified_with_empty(self) -> None:
        """Testing CertificateManager.mark_certificate_verified with empty
        fingerprints
        """
        cert_manager = CertificateManager()

        message = 'One or more SSL certificate fingerprints must be provided.'

        with self.assertRaisesMessage(ValueError, message):
            cert_manager.mark_certificate_verified(
                hostname='example.com',
                port=443,
                fingerprints=CertificateFingerprints())

    def test_remove_certificate_verification(self) -> None:
        """Testing CertificateManager.remove_certificate_verification"""
        cert_manager = CertificateManager()
        storage_backend = cert_manager.storage_backend
        fingerprints_path = os.path.join(storage_backend.storage_path,
                                         'fingerprints',
                                         'example.com__443.json')

        cert_manager.mark_certificate_verified(
            hostname='example.com',
            port=443,
            fingerprints=self.create_certificate_fingerprints())
        cert_manager.remove_certificate_verification(
            hostname='example.com',
            port=443)

        self.assertFalse(os.path.exists(fingerprints_path))

        # This should no longer be in cache.
        self.assertIsNone(cache.get(make_cache_key(
            'rb-ssl-fingerprints:file:example.com:443')))

    def test_remove_certificate_verification_with_local_site(self) -> None:
        """Testing CertificateManager.remove_certificate_verification with
        LocalSite
        """
        local_site = self.create_local_site('test-site-1')

        cert_manager = CertificateManager()
        storage_backend = cert_manager.storage_backend
        fingerprints_path = os.path.join(storage_backend.storage_path,
                                         'sites', 'test-site-1',
                                         'fingerprints',
                                         'example.com__443.json')

        cert_manager.mark_certificate_verified(
            hostname='example.com',
            port=443,
            fingerprints=self.create_certificate_fingerprints(),
            local_site=local_site)
        cert_manager.remove_certificate_verification(
            hostname='example.com',
            port=443,
            local_site=local_site)

        self.assertFalse(os.path.exists(fingerprints_path))

        # This should no longer be in cache.
        self.assertIsNone(cache.get(make_cache_key(
            'rb-ssl-fingerprints:file:1:example.com:443')))

    def test_remove_certificate_verification_with_not_found(self) -> None:
        """Testing CertificateManager.remove_certificate_verification with
        fingerprints not found
        """
        cert_manager = CertificateManager()

        # This should not blow up.
        cert_manager.remove_certificate_verification(
            hostname='missing.example.com',
            port=443)

    def test_remove_certificate_verification_with_local_site_mismatch(
        self,
    ) -> None:
        """Testing CertificateManager.remove_certificate_verification with
        LocalSite mismatch
        """
        local_site1 = self.create_local_site('test-site-1')
        local_site2 = self.create_local_site('test-site-2')

        cert_manager = CertificateManager()
        storage_backend = cert_manager.storage_backend
        fingerprints_path = os.path.join(storage_backend.storage_path,
                                         'sites', 'test-site-1',
                                         'fingerprints',
                                         'example.com__443.json')

        cert_manager.mark_certificate_verified(
            hostname='example.com',
            port=443,
            fingerprints=self.create_certificate_fingerprints(),
            local_site=local_site1)
        cert_manager.remove_certificate_verification(
            hostname='example.com',
            port=443,
            local_site=local_site2)

        self.assertTrue(os.path.exists(fingerprints_path))

        # This should still be in cache.
        self.assertEqual(
            cache.get(make_cache_key(
                'rb-ssl-fingerprints:file:1:example.com:443')),
            {
                'sha1': TEST_SHA1,
                'sha256': TEST_SHA256,
            })

    def test_get_verified_fingerprints(self) -> None:
        """Testing CertificateManager.get_verified_fingerprints"""
        local_site = self.create_local_site('test-site-1')

        cert_manager = CertificateManager()

        cert_manager.mark_certificate_verified(
            hostname='example.com',
            port=443,
            fingerprints=self.create_certificate_fingerprints())

        fingerprints = cert_manager.get_verified_fingerprints(
            hostname='example.com',
            port=443)

        self.assertAttrsEqual(
            fingerprints,
            {
                'sha1': TEST_SHA1,
                'sha256': TEST_SHA256,
            })

        # Make sure it's only accessible on the global site.
        self.assertIsNone(cert_manager.get_verified_fingerprints(
            hostname='example.com',
            port=443,
            local_site=local_site))

    def test_get_verified_fingerprints_with_local_site(self) -> None:
        """Testing CertificateManager.get_verified_fingerprints with LocalSite
        """
        local_site1 = self.create_local_site('test-site-1')
        local_site2 = self.create_local_site('test-site-2')

        cert_manager = CertificateManager()

        cert_manager.mark_certificate_verified(
            hostname='example.com',
            port=443,
            fingerprints=self.create_certificate_fingerprints(),
            local_site=local_site1)

        fingerprints = cert_manager.get_verified_fingerprints(
            hostname='example.com',
            port=443,
            local_site=local_site1)

        self.assertAttrsEqual(
            fingerprints,
            {
                'sha1': TEST_SHA1,
                'sha256': TEST_SHA256,
            })

        # Make sure it's only accessible on local_site1.
        self.assertIsNone(cert_manager.get_verified_fingerprints(
            hostname='example.com',
            port=443))
        self.assertIsNone(cert_manager.get_verified_fingerprints(
            hostname='example.com',
            port=443,
            local_site=local_site2))

    def test_get_verified_fingerprints_with_not_found(self) -> None:
        """Testing CertificateManager.get_verified_fingerprints with not found
        """
        cert_manager = CertificateManager()

        self.assertIsNone(cert_manager.get_verified_fingerprints(
            hostname='missing.example.com',
            port=443))

    def test_get_verified_fingerprints_after_unverify(self) -> None:
        """Testing CertificateManager.get_verified_fingerprints after
        remove_certificate_verification
        """
        cert_manager = CertificateManager()

        cert_manager.mark_certificate_verified(
            hostname='example.com',
            port=443,
            fingerprints=self.create_certificate_fingerprints())
        cert_manager.remove_certificate_verification(
            hostname='example.com',
            port=443)

        self.assertIsNone(cert_manager.get_verified_fingerprints(
            hostname='example.com',
            port=443))

    def test_is_certificate_verified_with_match(self) -> None:
        """Testing CertificateManager.is_certificate_verified with match"""
        cert_manager = CertificateManager()

        cert_manager.mark_certificate_verified(
            hostname='example.com',
            port=443,
            fingerprints=self.create_certificate_fingerprints())

        self.assertTrue(cert_manager.is_certificate_verified(
            hostname='example.com',
            port=443,
            latest_fingerprints=self.create_certificate_fingerprints()))

    def test_is_certificate_verified_with_no_match(self) -> None:
        """Testing CertificateManager.is_certificate_verified with no match"""
        local_site = self.create_local_site('test-site-1')

        cert_manager = CertificateManager()

        cert_manager.mark_certificate_verified(
            hostname='example.com',
            port=443,
            fingerprints=self.create_certificate_fingerprints())

        self.assertFalse(cert_manager.is_certificate_verified(
            hostname='example.com',
            port=443,
            latest_fingerprints=self.create_certificate_fingerprints(
                sha256=TEST_SHA256_2)))
        self.assertFalse(cert_manager.is_certificate_verified(
            hostname='example.com',
            port=443,
            latest_fingerprints=self.create_certificate_fingerprints(),
            local_site=local_site))

    def test_is_certificate_verified_after_unverify(self) -> None:
        """Testing CertificateManager.is_certificate_verified after
        remove_certificate_verification
        """
        cert_manager = CertificateManager()

        cert_manager.mark_certificate_verified(
            hostname='example.com',
            port=443,
            fingerprints=self.create_certificate_fingerprints())
        cert_manager.remove_certificate_verification(
            hostname='example.com',
            port=443)

        self.assertFalse(cert_manager.is_certificate_verified(
            hostname='example.com',
            port=443,
            latest_fingerprints=self.create_certificate_fingerprints()))

    def test_build_ssl_context(self) -> None:
        """Testing CertificateManager.build_ssl_context with defaults"""
        cert_manager = CertificateManager()
        storage_backend = cert_manager.storage_backend

        self.spy_on(ssl.create_default_context,
                    op=kgb.SpyOpReturn(CaptureSSLContext()))

        context = cert_manager.build_ssl_context(hostname='example.com',
                                                 port=443)
        assert isinstance(context, CaptureSSLContext)

        self.assertAttrsEqual(
            context,
            {
                'cadatas': [],
                'cafiles': [],
                'capaths': [
                    os.path.join(storage_backend.storage_path, 'cabundles'),
                ],
                'certfiles': [],
                'keyfiles': [],
                'passwords': [],
            })

    def test_build_ssl_context_with_ca_bundle(self) -> None:
        """Testing CertificateManager.build_ssl_context with CA bundles"""
        cert_manager = CertificateManager()
        storage_backend = cert_manager.storage_backend

        self.spy_on(ssl.create_default_context,
                    op=kgb.SpyOpReturn(CaptureSSLContext()))

        cert_manager.add_ca_bundle(
            CertificateBundle(bundle_data=TEST_CERT_BUNDLE_PEM,
                              name='my-certs'))

        context = cert_manager.build_ssl_context(hostname='example.com',
                                                 port=443)
        assert isinstance(context, CaptureSSLContext)

        self.assertAttrsEqual(
            context,
            {
                'cadatas': [],
                'cafiles': [],
                'capaths': [
                    os.path.join(storage_backend.storage_path, 'cabundles'),
                ],
                'certfiles': [],
                'keyfiles': [],
                'passwords': [],
            })

    def test_build_ssl_context_with_trust_cert(self) -> None:
        """Testing CertificateManager.build_ssl_context with trust
        certificate
        """
        cert_manager = CertificateManager()
        storage_backend = cert_manager.storage_backend

        self.spy_on(ssl.create_default_context,
                    op=kgb.SpyOpReturn(CaptureSSLContext()))

        cert_manager.add_certificate(
            self.create_certificate(cert_data=TEST_TRUST_CERT_PEM))

        context = cert_manager.build_ssl_context(hostname='example.com',
                                                 port=443)
        assert isinstance(context, CaptureSSLContext)

        self.assertAttrsEqual(
            context,
            {
                'cadatas': [],
                'cafiles': [
                    os.path.join(storage_backend.storage_path, 'certs',
                                 'trust', 'example.com__443.crt'),
                ],
                'capaths': [
                    os.path.join(storage_backend.storage_path, 'cabundles'),
                ],
                'certfiles': [],
                'keyfiles': [],
                'passwords': [],
            })

    def test_build_ssl_context_with_client_cert(self) -> None:
        """Testing CertificateManager.build_ssl_context with client
        certificate
        """
        cert_manager = CertificateManager()
        storage_backend = cert_manager.storage_backend

        self.spy_on(ssl.create_default_context,
                    op=kgb.SpyOpReturn(CaptureSSLContext()))

        cert_manager.add_certificate(
            self.create_certificate(purpose=CertPurpose.CLIENT,
                                    cert_data=TEST_CLIENT_CERT_PEM,
                                    key_data=TEST_CLIENT_KEY_PEM))

        context = cert_manager.build_ssl_context(hostname='example.com',
                                                 port=443)
        assert isinstance(context, CaptureSSLContext)

        self.assertAttrsEqual(
            context,
            {
                'cadatas': [],
                'cafiles': [],
                'capaths': [
                    os.path.join(storage_backend.storage_path, 'cabundles'),
                ],
                'certfiles': [
                    os.path.join(storage_backend.storage_path, 'certs',
                                 'client', 'example.com__443.crt'),
                ],
                'keyfiles': [
                    os.path.join(storage_backend.storage_path, 'certs',
                                 'client', 'example.com__443.key'),
                ],
                'passwords': [
                    None,
                ],
            })

    def test_build_ssl_context_with_all(self) -> None:
        """Testing CertificateManager.build_ssl_context with CA bundles,
        trust certificates, and client certificates
        """
        cert_manager = CertificateManager()
        storage_backend = cert_manager.storage_backend

        self.spy_on(ssl.create_default_context,
                    op=kgb.SpyOpReturn(CaptureSSLContext()))

        cert_manager.add_ca_bundle(
            CertificateBundle(bundle_data=TEST_CERT_BUNDLE_PEM,
                              name='my-certs'))
        cert_manager.add_certificate(
            self.create_certificate(cert_data=TEST_TRUST_CERT_PEM))
        cert_manager.add_certificate(
            self.create_certificate(purpose=CertPurpose.CLIENT,
                                    cert_data=TEST_CLIENT_CERT_PEM,
                                    key_data=TEST_CLIENT_KEY_PEM))

        context = cert_manager.build_ssl_context(hostname='example.com',
                                                 port=443)
        assert isinstance(context, CaptureSSLContext)

        self.assertAttrsEqual(
            context,
            {
                'cadatas': [],
                'cafiles': [
                    os.path.join(storage_backend.storage_path,
                                 'certs', 'trust', 'example.com__443.crt'),
                ],
                'capaths': [
                    os.path.join(storage_backend.storage_path, 'cabundles'),
                ],
                'certfiles': [
                    os.path.join(storage_backend.storage_path, 'certs',
                                 'client', 'example.com__443.crt'),
                ],
                'keyfiles': [
                    os.path.join(storage_backend.storage_path, 'certs',
                                 'client', 'example.com__443.key'),
                ],
                'passwords': [
                    None,
                ],
            })

    def test_build_ssl_context_with_local_site(self) -> None:
        """Testing CertificateManager.build_ssl_context with LocalSite"""
        local_site = self.create_local_site('test-site-1')

        cert_manager = CertificateManager()
        storage_backend = cert_manager.storage_backend
        site_storage_path = os.path.join(storage_backend.storage_path, 'sites',
                                         'test-site-1')

        self.spy_on(ssl.create_default_context,
                    op=kgb.SpyOpReturn(CaptureSSLContext()))

        cert_manager.add_certificate(
            self.create_certificate(cert_data=TEST_TRUST_CERT_PEM),
            local_site=local_site,
        )
        cert_manager.add_certificate(
            self.create_certificate(
                purpose=CertPurpose.CLIENT,
                cert_data=TEST_CLIENT_CERT_PEM,
                key_data=TEST_CLIENT_KEY_PEM,
            ),
            local_site=local_site,
        )

        # These should not be included.
        cert_manager.add_certificate(
            self.create_certificate(cert_data=TEST_TRUST_CERT_PEM),
        )
        cert_manager.add_certificate(
            self.create_certificate(
                purpose=CertPurpose.CLIENT,
                cert_data=TEST_CLIENT_CERT_PEM,
                key_data=TEST_CLIENT_KEY_PEM,
            ),
        )

        context = cert_manager.build_ssl_context(hostname='example.com',
                                                 port=443,
                                                 local_site=local_site)
        assert isinstance(context, CaptureSSLContext)

        self.assertAttrsEqual(
            context,
            {
                'cadatas': [],
                'cafiles': [
                    os.path.join(site_storage_path, 'certs', 'trust',
                                 'example.com__443.crt'),
                ],
                'capaths': [
                    os.path.join(site_storage_path, 'cabundles'),
                ],
                'certfiles': [
                    os.path.join(site_storage_path, 'certs', 'client',
                                 'example.com__443.crt'),
                ],
                'keyfiles': [
                    os.path.join(site_storage_path, 'certs', 'client',
                                 'example.com__443.key'),
                ],
                'passwords': [
                    None,
                ],
            })

    def test_build_urlopen_kwargs(self) -> None:
        """Testing CertificateManager.build_urlopen_kwargs with trust
        certificate
        """
        cert_manager = CertificateManager()
        storage_backend = cert_manager.storage_backend

        self.spy_on(ssl.create_default_context,
                    op=kgb.SpyOpReturn(CaptureSSLContext()))

        cert_manager.add_certificate(
            self.create_certificate(cert_data=TEST_TRUST_CERT_PEM))

        with self.assertWarns(RemovedInReviewBoard10_0Warning):
            kwargs = cert_manager.build_urlopen_kwargs(
                url='https://example.com',
            )

        self.assertAttrsEqual(
            kwargs.get('context'),
            {
                'cadatas': [],
                'cafiles': [
                    os.path.join(storage_backend.storage_path, 'certs',
                                 'trust', 'example.com__443.crt'),
                ],
                'capaths': [
                    os.path.join(storage_backend.storage_path, 'cabundles'),
                ],
                'certfiles': [],
                'keyfiles': [],
                'passwords': [],
            })

    def test_build_urlopen_kwargs_with_client_cert(self) -> None:
        """Testing CertificateManager.build_urlopen_kwargs with client
        certificate
        """
        cert_manager = CertificateManager()
        storage_backend = cert_manager.storage_backend

        self.spy_on(ssl.create_default_context,
                    op=kgb.SpyOpReturn(CaptureSSLContext()))

        cert_manager.add_certificate(
            self.create_certificate(purpose=CertPurpose.CLIENT,
                                    cert_data=TEST_CLIENT_CERT_PEM,
                                    key_data=TEST_CLIENT_KEY_PEM))

        with self.assertWarns(RemovedInReviewBoard10_0Warning):
            kwargs = cert_manager.build_urlopen_kwargs(
                url='https://example.com',
            )

        self.assertAttrsEqual(
            kwargs.get('context'),
            {
                'cadatas': [],
                'cafiles': [],
                'capaths': [
                    os.path.join(storage_backend.storage_path, 'cabundles'),
                ],
                'certfiles': [
                    os.path.join(storage_backend.storage_path, 'certs',
                                 'client', 'example.com__443.crt'),
                ],
                'keyfiles': [
                    os.path.join(storage_backend.storage_path, 'certs',
                                 'client', 'example.com__443.key'),
                ],
                'passwords': [
                    None,
                ],
            })

    def test_build_urlopen_kwargs_with_port(self) -> None:
        """Testing CertificateManager.build_urlopen_kwargs with port"""
        cert_manager = CertificateManager()
        storage_backend = cert_manager.storage_backend

        self.spy_on(ssl.create_default_context,
                    op=kgb.SpyOpReturn(CaptureSSLContext()))

        cert_manager.add_certificate(
            self.create_certificate(cert_data=TEST_TRUST_CERT_PEM,
                                    port=8443))

        with self.assertWarns(RemovedInReviewBoard10_0Warning):
            kwargs = cert_manager.build_urlopen_kwargs(
                url='https://example.com:8443',
            )

        self.assertAttrsEqual(
            kwargs.get('context'),
            {
                'cadatas': [],
                'cafiles': [
                    os.path.join(storage_backend.storage_path, 'certs',
                                 'trust', 'example.com__8443.crt'),
                ],
                'capaths': [
                    os.path.join(storage_backend.storage_path, 'cabundles'),
                ],
                'certfiles': [],
                'keyfiles': [],
                'passwords': [],
            })

    def test_build_urlopen_kwargs_with_non_https(self) -> None:
        """Testing CertificateManager.build_urlopen_kwargs with port"""
        cert_manager = CertificateManager()

        self.spy_on(ssl.create_default_context,
                    op=kgb.SpyOpReturn(CaptureSSLContext()))

        with self.assertWarns(RemovedInReviewBoard10_0Warning):
            kwargs = cert_manager.build_urlopen_kwargs(
                url='http://example.com',
            )

        self.assertEqual(kwargs, {})



"""Base support for defining UI themes.

Version Added:
    7.0
"""

from __future__ import annotations

from enum import Enum
from typing import ClassVar, TYPE_CHECKING

if TYPE_CHECKING:
    from django.http import HttpRequest
    from typelets.django.strings import StrOrPromise


class InkColorScheme(Enum):
    """The Ink color scheme set by a theme.

    Version Added:
        7.0
    """

    #: A light color scheme.
    LIGHT = 'light'

    #: A dark color scheme.
    DARK = 'dark'

    #: A high-contrast color scheme.
    HIGH_CONTRAST = 'high-contrast'

    #: A color scheme adapting to current system settings.
    SYSTEM = 'system'


class BaseUITheme:
    """Base class for a UI theme.

    Version Added:
        7.0
    """

    #: The unique ID of the theme.
    #:
    #: This is used for registration and lookup purposes.
    #:
    #: These only need to be unique within other UI themes.
    #:
    #: Type:
    #:     str
    theme_id: ClassVar[str]

    #: The displayed name of the theme.
    #:
    #: This will be shown to the user when choosing themes. It should be
    #: localized.
    #:
    #: Type:
    #:     str
    name: ClassVar[StrOrPromise]

    #: The value for the ``data-ink-color-scheme=`` HTML attribute.
    #:
    #: This informs the CSS and JavaScript of the current color scheme.
    #:
    #: Type:
    #:     InkColorScheme
    ink_color_scheme: ClassVar[InkColorScheme]

    def get_html_attrs(
        self,
        *,
        request: HttpRequest,
    ) -> dict[str, str]:
        """Return HTML attributes to include on the root HTML element.

        By default, this sets ``data-theme`` to :py:attr:`theme_id` and
        ``data-ink-color-scheme`` to :py:attr:`ink_color_scheme``.

        Subclasses can override this to provide additional attributes.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

        Returns:
            dict:
            The HTML attributes to include.
        """
        return {
            'data-ink-color-scheme': self.ink_color_scheme.value,
            'data-theme': self.theme_id,
        }
